from libc.stdint cimport int64_t, uint64_t
from libc.stddef cimport ptrdiff_t


cdef extern from "renpygl.h":

    ctypedef unsigned int GLenum
    ctypedef unsigned char GLboolean
    ctypedef unsigned int GLbitfield
    ctypedef void GLvoid
    ctypedef signed char GLbyte
    ctypedef short GLshort
    ctypedef int GLint
    ctypedef int GLclampx
    ctypedef unsigned char GLubyte
    ctypedef unsigned short GLushort
    ctypedef unsigned int GLuint
    ctypedef int GLsizei
    ctypedef float GLfloat
    ctypedef float GLclampf
    ctypedef double GLdouble
    ctypedef double GLclampd
    ctypedef void *GLeglClientBufferEXT
    ctypedef void *GLeglImageOES
    ctypedef char GLchar
    ctypedef char GLcharARB
    ctypedef unsigned short GLhalfARB
    ctypedef unsigned short GLhalf
    ctypedef GLint GLfixed
    ctypedef ptrdiff_t GLintptr
    ctypedef ptrdiff_t GLsizeiptr
    ctypedef int64_t GLint64
    ctypedef uint64_t GLuint64
    ctypedef ptrdiff_t GLintptrARB
    ctypedef ptrdiff_t GLsizeiptrARB
    ctypedef int64_t GLint64EXT
    ctypedef uint64_t GLuint64EXT

    GLenum GL_FALSE
    GLenum GL_NONE
    GLenum GL_NO_ERROR
    GLenum GL_POINTS
    GLenum GL_ZERO
    GLenum GL_LINES
    GLenum GL_MAP_READ_BIT
    GLenum GL_ONE
    GLenum GL_TRUE
    GLenum GL_LINE_LOOP
    GLenum GL_MAP_WRITE_BIT
    GLenum GL_LINE_STRIP
    GLenum GL_MAP_INVALIDATE_RANGE_BIT
    GLenum GL_TRIANGLES
    GLenum GL_TRIANGLE_STRIP
    GLenum GL_TRIANGLE_FAN
    GLenum GL_MAP_INVALIDATE_BUFFER_BIT
    GLenum GL_MAP_FLUSH_EXPLICIT_BIT
    GLenum GL_MAP_UNSYNCHRONIZED_BIT
    GLenum GL_DEPTH_BUFFER_BIT
    GLenum GL_NEVER
    GLenum GL_LESS
    GLenum GL_EQUAL
    GLenum GL_LEQUAL
    GLenum GL_GREATER
    GLenum GL_NOTEQUAL
    GLenum GL_GEQUAL
    GLenum GL_ALWAYS
    GLenum GL_SRC_COLOR
    GLenum GL_ONE_MINUS_SRC_COLOR
    GLenum GL_SRC_ALPHA
    GLenum GL_ONE_MINUS_SRC_ALPHA
    GLenum GL_DST_ALPHA
    GLenum GL_ONE_MINUS_DST_ALPHA
    GLenum GL_DST_COLOR
    GLenum GL_ONE_MINUS_DST_COLOR
    GLenum GL_SRC_ALPHA_SATURATE
    GLenum GL_STENCIL_BUFFER_BIT
    GLenum GL_FRONT
    GLenum GL_BACK
    GLenum GL_FRONT_AND_BACK
    GLenum GL_INVALID_ENUM
    GLenum GL_INVALID_VALUE
    GLenum GL_INVALID_OPERATION
    GLenum GL_OUT_OF_MEMORY
    GLenum GL_INVALID_FRAMEBUFFER_OPERATION
    GLenum GL_CW
    GLenum GL_CCW
    GLenum GL_LINE_WIDTH
    GLenum GL_CULL_FACE
    GLenum GL_CULL_FACE_MODE
    GLenum GL_FRONT_FACE
    GLenum GL_DEPTH_RANGE
    GLenum GL_DEPTH_TEST
    GLenum GL_DEPTH_WRITEMASK
    GLenum GL_DEPTH_CLEAR_VALUE
    GLenum GL_DEPTH_FUNC
    GLenum GL_STENCIL_TEST
    GLenum GL_STENCIL_CLEAR_VALUE
    GLenum GL_STENCIL_FUNC
    GLenum GL_STENCIL_VALUE_MASK
    GLenum GL_STENCIL_FAIL
    GLenum GL_STENCIL_PASS_DEPTH_FAIL
    GLenum GL_STENCIL_PASS_DEPTH_PASS
    GLenum GL_STENCIL_REF
    GLenum GL_STENCIL_WRITEMASK
    GLenum GL_VIEWPORT
    GLenum GL_DITHER
    GLenum GL_BLEND
    GLenum GL_READ_BUFFER
    GLenum GL_SCISSOR_BOX
    GLenum GL_SCISSOR_TEST
    GLenum GL_COLOR_CLEAR_VALUE
    GLenum GL_COLOR_WRITEMASK
    GLenum GL_UNPACK_ROW_LENGTH
    GLenum GL_UNPACK_SKIP_ROWS
    GLenum GL_UNPACK_SKIP_PIXELS
    GLenum GL_UNPACK_ALIGNMENT
    GLenum GL_PACK_ROW_LENGTH
    GLenum GL_PACK_SKIP_ROWS
    GLenum GL_PACK_SKIP_PIXELS
    GLenum GL_PACK_ALIGNMENT
    GLenum GL_MAX_TEXTURE_SIZE
    GLenum GL_MAX_VIEWPORT_DIMS
    GLenum GL_SUBPIXEL_BITS
    GLenum GL_RED_BITS
    GLenum GL_GREEN_BITS
    GLenum GL_BLUE_BITS
    GLenum GL_ALPHA_BITS
    GLenum GL_DEPTH_BITS
    GLenum GL_STENCIL_BITS
    GLenum GL_TEXTURE_2D
    GLenum GL_DONT_CARE
    GLenum GL_FASTEST
    GLenum GL_NICEST
    GLenum GL_BYTE
    GLenum GL_UNSIGNED_BYTE
    GLenum GL_SHORT
    GLenum GL_UNSIGNED_SHORT
    GLenum GL_INT
    GLenum GL_UNSIGNED_INT
    GLenum GL_FLOAT
    GLenum GL_HALF_FLOAT
    GLenum GL_INVERT
    GLenum GL_TEXTURE
    GLenum GL_COLOR
    GLenum GL_DEPTH
    GLenum GL_STENCIL
    GLenum GL_DEPTH_COMPONENT
    GLenum GL_RED
    GLenum GL_GREEN
    GLenum GL_BLUE
    GLenum GL_ALPHA
    GLenum GL_RGB
    GLenum GL_RGBA
    GLenum GL_LUMINANCE
    GLenum GL_LUMINANCE_ALPHA
    GLenum GL_KEEP
    GLenum GL_REPLACE
    GLenum GL_INCR
    GLenum GL_DECR
    GLenum GL_VENDOR
    GLenum GL_RENDERER
    GLenum GL_VERSION
    GLenum GL_EXTENSIONS
    GLenum GL_NEAREST
    GLenum GL_LINEAR
    GLenum GL_NEAREST_MIPMAP_NEAREST
    GLenum GL_LINEAR_MIPMAP_NEAREST
    GLenum GL_NEAREST_MIPMAP_LINEAR
    GLenum GL_LINEAR_MIPMAP_LINEAR
    GLenum GL_TEXTURE_MAG_FILTER
    GLenum GL_TEXTURE_MIN_FILTER
    GLenum GL_TEXTURE_WRAP_S
    GLenum GL_TEXTURE_WRAP_T
    GLenum GL_REPEAT
    GLenum GL_POLYGON_OFFSET_UNITS
    GLenum GL_COLOR_BUFFER_BIT
    GLenum GL_CONSTANT_COLOR
    GLenum GL_ONE_MINUS_CONSTANT_COLOR
    GLenum GL_CONSTANT_ALPHA
    GLenum GL_ONE_MINUS_CONSTANT_ALPHA
    GLenum GL_BLEND_COLOR
    GLenum GL_FUNC_ADD
    GLenum GL_MIN
    GLenum GL_MAX
    GLenum GL_BLEND_EQUATION
    GLenum GL_BLEND_EQUATION_RGB
    GLenum GL_FUNC_SUBTRACT
    GLenum GL_FUNC_REVERSE_SUBTRACT
    GLenum GL_UNSIGNED_SHORT_4_4_4_4
    GLenum GL_UNSIGNED_SHORT_5_5_5_1
    GLenum GL_POLYGON_OFFSET_FILL
    GLenum GL_POLYGON_OFFSET_FACTOR
    GLenum GL_RGB8
    GLenum GL_RGBA4
    GLenum GL_RGB5_A1
    GLenum GL_RGBA8
    GLenum GL_RGB10_A2
    GLenum GL_TEXTURE_BINDING_2D
    GLenum GL_TEXTURE_BINDING_3D
    GLenum GL_UNPACK_SKIP_IMAGES
    GLenum GL_UNPACK_IMAGE_HEIGHT
    GLenum GL_TEXTURE_3D
    GLenum GL_TEXTURE_WRAP_R
    GLenum GL_MAX_3D_TEXTURE_SIZE
    GLenum GL_SAMPLE_ALPHA_TO_COVERAGE
    GLenum GL_SAMPLE_COVERAGE
    GLenum GL_SAMPLE_BUFFERS
    GLenum GL_SAMPLES
    GLenum GL_SAMPLE_COVERAGE_VALUE
    GLenum GL_SAMPLE_COVERAGE_INVERT
    GLenum GL_BLEND_DST_RGB
    GLenum GL_BLEND_SRC_RGB
    GLenum GL_BLEND_DST_ALPHA
    GLenum GL_BLEND_SRC_ALPHA
    GLenum GL_MAX_ELEMENTS_VERTICES
    GLenum GL_MAX_ELEMENTS_INDICES
    GLenum GL_CLAMP_TO_EDGE
    GLenum GL_TEXTURE_MIN_LOD
    GLenum GL_TEXTURE_MAX_LOD
    GLenum GL_TEXTURE_BASE_LEVEL
    GLenum GL_TEXTURE_MAX_LEVEL
    GLenum GL_GENERATE_MIPMAP_HINT
    GLenum GL_DEPTH_COMPONENT16
    GLenum GL_DEPTH_COMPONENT24
    GLenum GL_FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING
    GLenum GL_FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_RED_SIZE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_GREEN_SIZE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_BLUE_SIZE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE
    GLenum GL_FRAMEBUFFER_DEFAULT
    GLenum GL_FRAMEBUFFER_UNDEFINED
    GLenum GL_DEPTH_STENCIL_ATTACHMENT
    GLenum GL_MAJOR_VERSION
    GLenum GL_MINOR_VERSION
    GLenum GL_NUM_EXTENSIONS
    GLenum GL_RG
    GLenum GL_RG_INTEGER
    GLenum GL_R8
    GLenum GL_RG8
    GLenum GL_R16F
    GLenum GL_R32F
    GLenum GL_RG16F
    GLenum GL_RG32F
    GLenum GL_R8I
    GLenum GL_R8UI
    GLenum GL_R16I
    GLenum GL_R16UI
    GLenum GL_R32I
    GLenum GL_R32UI
    GLenum GL_RG8I
    GLenum GL_RG8UI
    GLenum GL_RG16I
    GLenum GL_RG16UI
    GLenum GL_RG32I
    GLenum GL_RG32UI
    GLenum GL_UNSIGNED_SHORT_5_6_5
    GLenum GL_UNSIGNED_INT_2_10_10_10_REV
    GLenum GL_MIRRORED_REPEAT
    GLenum GL_ALIASED_POINT_SIZE_RANGE
    GLenum GL_ALIASED_LINE_WIDTH_RANGE
    GLenum GL_TEXTURE0
    GLenum GL_TEXTURE1
    GLenum GL_TEXTURE2
    GLenum GL_TEXTURE3
    GLenum GL_TEXTURE4
    GLenum GL_TEXTURE5
    GLenum GL_TEXTURE6
    GLenum GL_TEXTURE7
    GLenum GL_TEXTURE8
    GLenum GL_TEXTURE9
    GLenum GL_TEXTURE10
    GLenum GL_TEXTURE11
    GLenum GL_TEXTURE12
    GLenum GL_TEXTURE13
    GLenum GL_TEXTURE14
    GLenum GL_TEXTURE15
    GLenum GL_TEXTURE16
    GLenum GL_TEXTURE17
    GLenum GL_TEXTURE18
    GLenum GL_TEXTURE19
    GLenum GL_TEXTURE20
    GLenum GL_TEXTURE21
    GLenum GL_TEXTURE22
    GLenum GL_TEXTURE23
    GLenum GL_TEXTURE24
    GLenum GL_TEXTURE25
    GLenum GL_TEXTURE26
    GLenum GL_TEXTURE27
    GLenum GL_TEXTURE28
    GLenum GL_TEXTURE29
    GLenum GL_TEXTURE30
    GLenum GL_TEXTURE31
    GLenum GL_ACTIVE_TEXTURE
    GLenum GL_MAX_RENDERBUFFER_SIZE
    GLenum GL_DEPTH_STENCIL
    GLenum GL_UNSIGNED_INT_24_8
    GLenum GL_MAX_TEXTURE_LOD_BIAS
    GLenum GL_INCR_WRAP
    GLenum GL_DECR_WRAP
    GLenum GL_TEXTURE_CUBE_MAP
    GLenum GL_TEXTURE_BINDING_CUBE_MAP
    GLenum GL_TEXTURE_CUBE_MAP_POSITIVE_X
    GLenum GL_TEXTURE_CUBE_MAP_NEGATIVE_X
    GLenum GL_TEXTURE_CUBE_MAP_POSITIVE_Y
    GLenum GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
    GLenum GL_TEXTURE_CUBE_MAP_POSITIVE_Z
    GLenum GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
    GLenum GL_MAX_CUBE_MAP_TEXTURE_SIZE
    GLenum GL_VERTEX_ARRAY_BINDING
    GLenum GL_VERTEX_ATTRIB_ARRAY_ENABLED
    GLenum GL_VERTEX_ATTRIB_ARRAY_SIZE
    GLenum GL_VERTEX_ATTRIB_ARRAY_STRIDE
    GLenum GL_VERTEX_ATTRIB_ARRAY_TYPE
    GLenum GL_CURRENT_VERTEX_ATTRIB
    GLenum GL_VERTEX_ATTRIB_ARRAY_POINTER
    GLenum GL_NUM_COMPRESSED_TEXTURE_FORMATS
    GLenum GL_COMPRESSED_TEXTURE_FORMATS
    GLenum GL_BUFFER_SIZE
    GLenum GL_BUFFER_USAGE
    GLenum GL_STENCIL_BACK_FUNC
    GLenum GL_STENCIL_BACK_FAIL
    GLenum GL_STENCIL_BACK_PASS_DEPTH_FAIL
    GLenum GL_STENCIL_BACK_PASS_DEPTH_PASS
    GLenum GL_RGBA32F
    GLenum GL_RGB32F
    GLenum GL_RGBA16F
    GLenum GL_RGB16F
    GLenum GL_MAX_DRAW_BUFFERS
    GLenum GL_DRAW_BUFFER0
    GLenum GL_DRAW_BUFFER1
    GLenum GL_DRAW_BUFFER2
    GLenum GL_DRAW_BUFFER3
    GLenum GL_DRAW_BUFFER4
    GLenum GL_DRAW_BUFFER5
    GLenum GL_DRAW_BUFFER6
    GLenum GL_DRAW_BUFFER7
    GLenum GL_DRAW_BUFFER8
    GLenum GL_DRAW_BUFFER9
    GLenum GL_DRAW_BUFFER10
    GLenum GL_DRAW_BUFFER11
    GLenum GL_DRAW_BUFFER12
    GLenum GL_DRAW_BUFFER13
    GLenum GL_DRAW_BUFFER14
    GLenum GL_DRAW_BUFFER15
    GLenum GL_BLEND_EQUATION_ALPHA
    GLenum GL_TEXTURE_COMPARE_MODE
    GLenum GL_TEXTURE_COMPARE_FUNC
    GLenum GL_COMPARE_REF_TO_TEXTURE
    GLenum GL_CURRENT_QUERY
    GLenum GL_QUERY_RESULT
    GLenum GL_QUERY_RESULT_AVAILABLE
    GLenum GL_MAX_VERTEX_ATTRIBS
    GLenum GL_VERTEX_ATTRIB_ARRAY_NORMALIZED
    GLenum GL_MAX_TEXTURE_IMAGE_UNITS
    GLenum GL_ARRAY_BUFFER
    GLenum GL_ELEMENT_ARRAY_BUFFER
    GLenum GL_ARRAY_BUFFER_BINDING
    GLenum GL_ELEMENT_ARRAY_BUFFER_BINDING
    GLenum GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING
    GLenum GL_BUFFER_MAPPED
    GLenum GL_BUFFER_MAP_POINTER
    GLenum GL_STREAM_DRAW
    GLenum GL_STREAM_READ
    GLenum GL_STREAM_COPY
    GLenum GL_STATIC_DRAW
    GLenum GL_STATIC_READ
    GLenum GL_STATIC_COPY
    GLenum GL_DYNAMIC_DRAW
    GLenum GL_DYNAMIC_READ
    GLenum GL_DYNAMIC_COPY
    GLenum GL_PIXEL_PACK_BUFFER
    GLenum GL_PIXEL_UNPACK_BUFFER
    GLenum GL_PIXEL_PACK_BUFFER_BINDING
    GLenum GL_PIXEL_UNPACK_BUFFER_BINDING
    GLenum GL_DEPTH24_STENCIL8
    GLenum GL_VERTEX_ATTRIB_ARRAY_INTEGER
    GLenum GL_MAX_ARRAY_TEXTURE_LAYERS
    GLenum GL_MIN_PROGRAM_TEXEL_OFFSET
    GLenum GL_MAX_PROGRAM_TEXEL_OFFSET
    GLenum GL_FRAGMENT_SHADER
    GLenum GL_VERTEX_SHADER
    GLenum GL_MAX_FRAGMENT_UNIFORM_COMPONENTS
    GLenum GL_MAX_VERTEX_UNIFORM_COMPONENTS
    GLenum GL_MAX_VARYING_COMPONENTS
    GLenum GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS
    GLenum GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS
    GLenum GL_SHADER_TYPE
    GLenum GL_FLOAT_VEC2
    GLenum GL_FLOAT_VEC3
    GLenum GL_FLOAT_VEC4
    GLenum GL_INT_VEC2
    GLenum GL_INT_VEC3
    GLenum GL_INT_VEC4
    GLenum GL_BOOL
    GLenum GL_BOOL_VEC2
    GLenum GL_BOOL_VEC3
    GLenum GL_BOOL_VEC4
    GLenum GL_FLOAT_MAT2
    GLenum GL_FLOAT_MAT3
    GLenum GL_FLOAT_MAT4
    GLenum GL_SAMPLER_2D
    GLenum GL_SAMPLER_3D
    GLenum GL_SAMPLER_CUBE
    GLenum GL_SAMPLER_2D_SHADOW
    GLenum GL_FLOAT_MAT2x3
    GLenum GL_FLOAT_MAT2x4
    GLenum GL_FLOAT_MAT3x2
    GLenum GL_FLOAT_MAT3x4
    GLenum GL_FLOAT_MAT4x2
    GLenum GL_FLOAT_MAT4x3
    GLenum GL_DELETE_STATUS
    GLenum GL_COMPILE_STATUS
    GLenum GL_LINK_STATUS
    GLenum GL_VALIDATE_STATUS
    GLenum GL_INFO_LOG_LENGTH
    GLenum GL_ATTACHED_SHADERS
    GLenum GL_ACTIVE_UNIFORMS
    GLenum GL_ACTIVE_UNIFORM_MAX_LENGTH
    GLenum GL_SHADER_SOURCE_LENGTH
    GLenum GL_ACTIVE_ATTRIBUTES
    GLenum GL_ACTIVE_ATTRIBUTE_MAX_LENGTH
    GLenum GL_FRAGMENT_SHADER_DERIVATIVE_HINT
    GLenum GL_SHADING_LANGUAGE_VERSION
    GLenum GL_CURRENT_PROGRAM
    GLenum GL_UNSIGNED_NORMALIZED
    GLenum GL_TEXTURE_2D_ARRAY
    GLenum GL_TEXTURE_BINDING_2D_ARRAY
    GLenum GL_R11F_G11F_B10F
    GLenum GL_UNSIGNED_INT_10F_11F_11F_REV
    GLenum GL_RGB9_E5
    GLenum GL_UNSIGNED_INT_5_9_9_9_REV
    GLenum GL_SRGB
    GLenum GL_SRGB8
    GLenum GL_SRGB8_ALPHA8
    GLenum GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH
    GLenum GL_TRANSFORM_FEEDBACK_BUFFER_MODE
    GLenum GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS
    GLenum GL_TRANSFORM_FEEDBACK_VARYINGS
    GLenum GL_TRANSFORM_FEEDBACK_BUFFER_START
    GLenum GL_TRANSFORM_FEEDBACK_BUFFER_SIZE
    GLenum GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN
    GLenum GL_RASTERIZER_DISCARD
    GLenum GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS
    GLenum GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS
    GLenum GL_INTERLEAVED_ATTRIBS
    GLenum GL_SEPARATE_ATTRIBS
    GLenum GL_TRANSFORM_FEEDBACK_BUFFER
    GLenum GL_TRANSFORM_FEEDBACK_BUFFER_BINDING
    GLenum GL_STENCIL_BACK_REF
    GLenum GL_STENCIL_BACK_VALUE_MASK
    GLenum GL_STENCIL_BACK_WRITEMASK
    GLenum GL_DRAW_FRAMEBUFFER_BINDING
    GLenum GL_FRAMEBUFFER_BINDING
    GLenum GL_RENDERBUFFER_BINDING
    GLenum GL_READ_FRAMEBUFFER
    GLenum GL_DRAW_FRAMEBUFFER
    GLenum GL_READ_FRAMEBUFFER_BINDING
    GLenum GL_RENDERBUFFER_SAMPLES
    GLenum GL_DEPTH_COMPONENT32F
    GLenum GL_DEPTH32F_STENCIL8
    GLenum GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME
    GLenum GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL
    GLenum GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE
    GLenum GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER
    GLenum GL_FRAMEBUFFER_COMPLETE
    GLenum GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT
    GLenum GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT
    GLenum GL_FRAMEBUFFER_UNSUPPORTED
    GLenum GL_MAX_COLOR_ATTACHMENTS
    GLenum GL_COLOR_ATTACHMENT0
    GLenum GL_COLOR_ATTACHMENT1
    GLenum GL_COLOR_ATTACHMENT2
    GLenum GL_COLOR_ATTACHMENT3
    GLenum GL_COLOR_ATTACHMENT4
    GLenum GL_COLOR_ATTACHMENT5
    GLenum GL_COLOR_ATTACHMENT6
    GLenum GL_COLOR_ATTACHMENT7
    GLenum GL_COLOR_ATTACHMENT8
    GLenum GL_COLOR_ATTACHMENT9
    GLenum GL_COLOR_ATTACHMENT10
    GLenum GL_COLOR_ATTACHMENT11
    GLenum GL_COLOR_ATTACHMENT12
    GLenum GL_COLOR_ATTACHMENT13
    GLenum GL_COLOR_ATTACHMENT14
    GLenum GL_COLOR_ATTACHMENT15
    GLenum GL_COLOR_ATTACHMENT16
    GLenum GL_COLOR_ATTACHMENT17
    GLenum GL_COLOR_ATTACHMENT18
    GLenum GL_COLOR_ATTACHMENT19
    GLenum GL_COLOR_ATTACHMENT20
    GLenum GL_COLOR_ATTACHMENT21
    GLenum GL_COLOR_ATTACHMENT22
    GLenum GL_COLOR_ATTACHMENT23
    GLenum GL_COLOR_ATTACHMENT24
    GLenum GL_COLOR_ATTACHMENT25
    GLenum GL_COLOR_ATTACHMENT26
    GLenum GL_COLOR_ATTACHMENT27
    GLenum GL_COLOR_ATTACHMENT28
    GLenum GL_COLOR_ATTACHMENT29
    GLenum GL_COLOR_ATTACHMENT30
    GLenum GL_COLOR_ATTACHMENT31
    GLenum GL_DEPTH_ATTACHMENT
    GLenum GL_STENCIL_ATTACHMENT
    GLenum GL_FRAMEBUFFER
    GLenum GL_RENDERBUFFER
    GLenum GL_RENDERBUFFER_WIDTH
    GLenum GL_RENDERBUFFER_HEIGHT
    GLenum GL_RENDERBUFFER_INTERNAL_FORMAT
    GLenum GL_STENCIL_INDEX8
    GLenum GL_RENDERBUFFER_RED_SIZE
    GLenum GL_RENDERBUFFER_GREEN_SIZE
    GLenum GL_RENDERBUFFER_BLUE_SIZE
    GLenum GL_RENDERBUFFER_ALPHA_SIZE
    GLenum GL_RENDERBUFFER_DEPTH_SIZE
    GLenum GL_RENDERBUFFER_STENCIL_SIZE
    GLenum GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE
    GLenum GL_MAX_SAMPLES
    GLenum GL_RGBA32UI
    GLenum GL_RGB32UI
    GLenum GL_RGBA16UI
    GLenum GL_RGB16UI
    GLenum GL_RGBA8UI
    GLenum GL_RGB8UI
    GLenum GL_RGBA32I
    GLenum GL_RGB32I
    GLenum GL_RGBA16I
    GLenum GL_RGB16I
    GLenum GL_RGBA8I
    GLenum GL_RGB8I
    GLenum GL_RED_INTEGER
    GLenum GL_RGB_INTEGER
    GLenum GL_RGBA_INTEGER
    GLenum GL_FLOAT_32_UNSIGNED_INT_24_8_REV
    GLenum GL_SAMPLER_2D_ARRAY
    GLenum GL_SAMPLER_2D_ARRAY_SHADOW
    GLenum GL_SAMPLER_CUBE_SHADOW
    GLenum GL_UNSIGNED_INT_VEC2
    GLenum GL_UNSIGNED_INT_VEC3
    GLenum GL_UNSIGNED_INT_VEC4
    GLenum GL_INT_SAMPLER_2D
    GLenum GL_INT_SAMPLER_3D
    GLenum GL_INT_SAMPLER_CUBE
    GLenum GL_INT_SAMPLER_2D_ARRAY
    GLenum GL_UNSIGNED_INT_SAMPLER_2D
    GLenum GL_UNSIGNED_INT_SAMPLER_3D
    GLenum GL_UNSIGNED_INT_SAMPLER_CUBE
    GLenum GL_UNSIGNED_INT_SAMPLER_2D_ARRAY
    GLenum GL_BUFFER_ACCESS_FLAGS
    GLenum GL_BUFFER_MAP_LENGTH
    GLenum GL_BUFFER_MAP_OFFSET

ctypedef void (__stdcall *glActiveTexture_type)(GLenum  texture) nogil
cdef glActiveTexture_type glActiveTexture

ctypedef void (__stdcall *glAttachShader_type)(GLuint  program, GLuint  shader) nogil
cdef glAttachShader_type glAttachShader

ctypedef void (__stdcall *glBeginQuery_type)(GLenum  target, GLuint  id) nogil
cdef glBeginQuery_type glBeginQuery

ctypedef void (__stdcall *glBeginTransformFeedback_type)(GLenum  primitiveMode) nogil
cdef glBeginTransformFeedback_type glBeginTransformFeedback

ctypedef void (__stdcall *glBindAttribLocation_type)(GLuint  program, GLuint  index, const GLchar * name) nogil
cdef glBindAttribLocation_type glBindAttribLocation

ctypedef void (__stdcall *glBindBuffer_type)(GLenum  target, GLuint  buffer) nogil
cdef glBindBuffer_type glBindBuffer

ctypedef void (__stdcall *glBindBufferBase_type)(GLenum  target, GLuint  index, GLuint  buffer) nogil
cdef glBindBufferBase_type glBindBufferBase

ctypedef void (__stdcall *glBindBufferRange_type)(GLenum  target, GLuint  index, GLuint  buffer, GLintptr  offset, GLsizeiptr  size) nogil
cdef glBindBufferRange_type glBindBufferRange

ctypedef void (__stdcall *glBindFramebuffer_type)(GLenum  target, GLuint  framebuffer) nogil
cdef glBindFramebuffer_type glBindFramebuffer

ctypedef void (__stdcall *glBindRenderbuffer_type)(GLenum  target, GLuint  renderbuffer) nogil
cdef glBindRenderbuffer_type glBindRenderbuffer

ctypedef void (__stdcall *glBindTexture_type)(GLenum  target, GLuint  texture) nogil
cdef glBindTexture_type glBindTexture

ctypedef void (__stdcall *glBindVertexArray_type)(GLuint  array) nogil
cdef glBindVertexArray_type glBindVertexArray

ctypedef void (__stdcall *glBlendColor_type)(GLfloat  red, GLfloat  green, GLfloat  blue, GLfloat  alpha) nogil
cdef glBlendColor_type glBlendColor

ctypedef void (__stdcall *glBlendEquation_type)(GLenum  mode) nogil
cdef glBlendEquation_type glBlendEquation

ctypedef void (__stdcall *glBlendEquationSeparate_type)(GLenum  modeRGB, GLenum  modeAlpha) nogil
cdef glBlendEquationSeparate_type glBlendEquationSeparate

ctypedef void (__stdcall *glBlendFunc_type)(GLenum  sfactor, GLenum  dfactor) nogil
cdef glBlendFunc_type glBlendFunc

ctypedef void (__stdcall *glBlendFuncSeparate_type)(GLenum  sfactorRGB, GLenum  dfactorRGB, GLenum  sfactorAlpha, GLenum  dfactorAlpha) nogil
cdef glBlendFuncSeparate_type glBlendFuncSeparate

ctypedef void (__stdcall *glBlitFramebuffer_type)(GLint  srcX0, GLint  srcY0, GLint  srcX1, GLint  srcY1, GLint  dstX0, GLint  dstY0, GLint  dstX1, GLint  dstY1, GLbitfield  mask, GLenum  filter) nogil
cdef glBlitFramebuffer_type glBlitFramebuffer

ctypedef void (__stdcall *glBufferData_type)(GLenum  target, GLsizeiptr  size, const void * data, GLenum  usage) nogil
cdef glBufferData_type glBufferData

ctypedef void (__stdcall *glBufferSubData_type)(GLenum  target, GLintptr  offset, GLsizeiptr  size, const void * data) nogil
cdef glBufferSubData_type glBufferSubData

ctypedef GLenum (__stdcall *glCheckFramebufferStatus_type)(GLenum  target) nogil
cdef glCheckFramebufferStatus_type glCheckFramebufferStatus

ctypedef void (__stdcall *glClear_type)(GLbitfield  mask) nogil
cdef glClear_type glClear

ctypedef void (__stdcall *glClearBufferfi_type)(GLenum  buffer, GLint  drawbuffer, GLfloat  depth, GLint  stencil) nogil
cdef glClearBufferfi_type glClearBufferfi

ctypedef void (__stdcall *glClearBufferfv_type)(GLenum  buffer, GLint  drawbuffer, const GLfloat * value) nogil
cdef glClearBufferfv_type glClearBufferfv

ctypedef void (__stdcall *glClearBufferiv_type)(GLenum  buffer, GLint  drawbuffer, const GLint * value) nogil
cdef glClearBufferiv_type glClearBufferiv

ctypedef void (__stdcall *glClearBufferuiv_type)(GLenum  buffer, GLint  drawbuffer, const GLuint * value) nogil
cdef glClearBufferuiv_type glClearBufferuiv

ctypedef void (__stdcall *glClearColor_type)(GLfloat  red, GLfloat  green, GLfloat  blue, GLfloat  alpha) nogil
cdef glClearColor_type glClearColor

ctypedef void (__stdcall *glClearStencil_type)(GLint  s) nogil
cdef glClearStencil_type glClearStencil

ctypedef void (__stdcall *glColorMask_type)(GLboolean  red, GLboolean  green, GLboolean  blue, GLboolean  alpha) nogil
cdef glColorMask_type glColorMask

ctypedef void (__stdcall *glCompileShader_type)(GLuint  shader) nogil
cdef glCompileShader_type glCompileShader

ctypedef void (__stdcall *glCompressedTexImage2D_type)(GLenum  target, GLint  level, GLenum  internalformat, GLsizei  width, GLsizei  height, GLint  border, GLsizei  imageSize, const void * data) nogil
cdef glCompressedTexImage2D_type glCompressedTexImage2D

ctypedef void (__stdcall *glCompressedTexImage3D_type)(GLenum  target, GLint  level, GLenum  internalformat, GLsizei  width, GLsizei  height, GLsizei  depth, GLint  border, GLsizei  imageSize, const void * data) nogil
cdef glCompressedTexImage3D_type glCompressedTexImage3D

ctypedef void (__stdcall *glCompressedTexSubImage2D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLsizei  width, GLsizei  height, GLenum  format, GLsizei  imageSize, const void * data) nogil
cdef glCompressedTexSubImage2D_type glCompressedTexSubImage2D

ctypedef void (__stdcall *glCompressedTexSubImage3D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLint  zoffset, GLsizei  width, GLsizei  height, GLsizei  depth, GLenum  format, GLsizei  imageSize, const void * data) nogil
cdef glCompressedTexSubImage3D_type glCompressedTexSubImage3D

ctypedef void (__stdcall *glCopyTexImage2D_type)(GLenum  target, GLint  level, GLenum  internalformat, GLint  x, GLint  y, GLsizei  width, GLsizei  height, GLint  border) nogil
cdef glCopyTexImage2D_type glCopyTexImage2D

ctypedef void (__stdcall *glCopyTexSubImage2D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLint  x, GLint  y, GLsizei  width, GLsizei  height) nogil
cdef glCopyTexSubImage2D_type glCopyTexSubImage2D

ctypedef void (__stdcall *glCopyTexSubImage3D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLint  zoffset, GLint  x, GLint  y, GLsizei  width, GLsizei  height) nogil
cdef glCopyTexSubImage3D_type glCopyTexSubImage3D

ctypedef GLuint (__stdcall *glCreateProgram_type)() nogil
cdef glCreateProgram_type glCreateProgram

ctypedef GLuint (__stdcall *glCreateShader_type)(GLenum  type) nogil
cdef glCreateShader_type glCreateShader

ctypedef void (__stdcall *glCullFace_type)(GLenum  mode) nogil
cdef glCullFace_type glCullFace

ctypedef void (__stdcall *glDeleteBuffers_type)(GLsizei  n, const GLuint * buffers) nogil
cdef glDeleteBuffers_type glDeleteBuffers

ctypedef void (__stdcall *glDeleteFramebuffers_type)(GLsizei  n, const GLuint * framebuffers) nogil
cdef glDeleteFramebuffers_type glDeleteFramebuffers

ctypedef void (__stdcall *glDeleteProgram_type)(GLuint  program) nogil
cdef glDeleteProgram_type glDeleteProgram

ctypedef void (__stdcall *glDeleteQueries_type)(GLsizei  n, const GLuint * ids) nogil
cdef glDeleteQueries_type glDeleteQueries

ctypedef void (__stdcall *glDeleteRenderbuffers_type)(GLsizei  n, const GLuint * renderbuffers) nogil
cdef glDeleteRenderbuffers_type glDeleteRenderbuffers

ctypedef void (__stdcall *glDeleteShader_type)(GLuint  shader) nogil
cdef glDeleteShader_type glDeleteShader

ctypedef void (__stdcall *glDeleteTextures_type)(GLsizei  n, const GLuint * textures) nogil
cdef glDeleteTextures_type glDeleteTextures

ctypedef void (__stdcall *glDeleteVertexArrays_type)(GLsizei  n, const GLuint * arrays) nogil
cdef glDeleteVertexArrays_type glDeleteVertexArrays

ctypedef void (__stdcall *glDepthFunc_type)(GLenum  func) nogil
cdef glDepthFunc_type glDepthFunc

ctypedef void (__stdcall *glDepthMask_type)(GLboolean  flag) nogil
cdef glDepthMask_type glDepthMask

ctypedef void (__stdcall *glDetachShader_type)(GLuint  program, GLuint  shader) nogil
cdef glDetachShader_type glDetachShader

ctypedef void (__stdcall *glDisable_type)(GLenum  cap) nogil
cdef glDisable_type glDisable

ctypedef void (__stdcall *glDisableVertexAttribArray_type)(GLuint  index) nogil
cdef glDisableVertexAttribArray_type glDisableVertexAttribArray

ctypedef void (__stdcall *glDrawArrays_type)(GLenum  mode, GLint  first, GLsizei  count) nogil
cdef glDrawArrays_type glDrawArrays

ctypedef void (__stdcall *glDrawBuffers_type)(GLsizei  n, const GLenum * bufs) nogil
cdef glDrawBuffers_type glDrawBuffers

ctypedef void (__stdcall *glDrawElements_type)(GLenum  mode, GLsizei  count, GLenum  type, const void * indices) nogil
cdef glDrawElements_type glDrawElements

ctypedef void (__stdcall *glDrawRangeElements_type)(GLenum  mode, GLuint  start, GLuint  end, GLsizei  count, GLenum  type, const void * indices) nogil
cdef glDrawRangeElements_type glDrawRangeElements

ctypedef void (__stdcall *glEnable_type)(GLenum  cap) nogil
cdef glEnable_type glEnable

ctypedef void (__stdcall *glEnableVertexAttribArray_type)(GLuint  index) nogil
cdef glEnableVertexAttribArray_type glEnableVertexAttribArray

ctypedef void (__stdcall *glEndQuery_type)(GLenum  target) nogil
cdef glEndQuery_type glEndQuery

ctypedef void (__stdcall *glEndTransformFeedback_type)() nogil
cdef glEndTransformFeedback_type glEndTransformFeedback

ctypedef void (__stdcall *glFinish_type)() nogil
cdef glFinish_type glFinish

ctypedef void (__stdcall *glFlush_type)() nogil
cdef glFlush_type glFlush

ctypedef void (__stdcall *glFlushMappedBufferRange_type)(GLenum  target, GLintptr  offset, GLsizeiptr  length) nogil
cdef glFlushMappedBufferRange_type glFlushMappedBufferRange

ctypedef void (__stdcall *glFramebufferRenderbuffer_type)(GLenum  target, GLenum  attachment, GLenum  renderbuffertarget, GLuint  renderbuffer) nogil
cdef glFramebufferRenderbuffer_type glFramebufferRenderbuffer

ctypedef void (__stdcall *glFramebufferTexture2D_type)(GLenum  target, GLenum  attachment, GLenum  textarget, GLuint  texture, GLint  level) nogil
cdef glFramebufferTexture2D_type glFramebufferTexture2D

ctypedef void (__stdcall *glFramebufferTextureLayer_type)(GLenum  target, GLenum  attachment, GLuint  texture, GLint  level, GLint  layer) nogil
cdef glFramebufferTextureLayer_type glFramebufferTextureLayer

ctypedef void (__stdcall *glFrontFace_type)(GLenum  mode) nogil
cdef glFrontFace_type glFrontFace

ctypedef void (__stdcall *glGenBuffers_type)(GLsizei  n, GLuint * buffers) nogil
cdef glGenBuffers_type glGenBuffers

ctypedef void (__stdcall *glGenFramebuffers_type)(GLsizei  n, GLuint * framebuffers) nogil
cdef glGenFramebuffers_type glGenFramebuffers

ctypedef void (__stdcall *glGenQueries_type)(GLsizei  n, GLuint * ids) nogil
cdef glGenQueries_type glGenQueries

ctypedef void (__stdcall *glGenRenderbuffers_type)(GLsizei  n, GLuint * renderbuffers) nogil
cdef glGenRenderbuffers_type glGenRenderbuffers

ctypedef void (__stdcall *glGenTextures_type)(GLsizei  n, GLuint * textures) nogil
cdef glGenTextures_type glGenTextures

ctypedef void (__stdcall *glGenVertexArrays_type)(GLsizei  n, GLuint * arrays) nogil
cdef glGenVertexArrays_type glGenVertexArrays

ctypedef void (__stdcall *glGenerateMipmap_type)(GLenum  target) nogil
cdef glGenerateMipmap_type glGenerateMipmap

ctypedef void (__stdcall *glGetActiveAttrib_type)(GLuint  program, GLuint  index, GLsizei  bufSize, GLsizei * length, GLint * size, GLenum * type, GLchar * name) nogil
cdef glGetActiveAttrib_type glGetActiveAttrib

ctypedef void (__stdcall *glGetActiveUniform_type)(GLuint  program, GLuint  index, GLsizei  bufSize, GLsizei * length, GLint * size, GLenum * type, GLchar * name) nogil
cdef glGetActiveUniform_type glGetActiveUniform

ctypedef void (__stdcall *glGetAttachedShaders_type)(GLuint  program, GLsizei  maxCount, GLsizei * count, GLuint * shaders) nogil
cdef glGetAttachedShaders_type glGetAttachedShaders

ctypedef GLint (__stdcall *glGetAttribLocation_type)(GLuint  program, const GLchar * name) nogil
cdef glGetAttribLocation_type glGetAttribLocation

ctypedef void (__stdcall *glGetBooleanv_type)(GLenum  pname, GLboolean * data) nogil
cdef glGetBooleanv_type glGetBooleanv

ctypedef void (__stdcall *glGetBufferParameteriv_type)(GLenum  target, GLenum  pname, GLint * params) nogil
cdef glGetBufferParameteriv_type glGetBufferParameteriv

ctypedef void (__stdcall *glGetBufferPointerv_type)(GLenum  target, GLenum  pname, void ** params) nogil
cdef glGetBufferPointerv_type glGetBufferPointerv

ctypedef GLenum (__stdcall *glGetError_type)() nogil
cdef glGetError_type glGetError

ctypedef void (__stdcall *glGetFloatv_type)(GLenum  pname, GLfloat * data) nogil
cdef glGetFloatv_type glGetFloatv

ctypedef GLint (__stdcall *glGetFragDataLocation_type)(GLuint  program, const GLchar * name) nogil
cdef glGetFragDataLocation_type glGetFragDataLocation

ctypedef void (__stdcall *glGetFramebufferAttachmentParameteriv_type)(GLenum  target, GLenum  attachment, GLenum  pname, GLint * params) nogil
cdef glGetFramebufferAttachmentParameteriv_type glGetFramebufferAttachmentParameteriv

ctypedef void (__stdcall *glGetIntegeri_v_type)(GLenum  target, GLuint  index, GLint * data) nogil
cdef glGetIntegeri_v_type glGetIntegeri_v

ctypedef void (__stdcall *glGetIntegerv_type)(GLenum  pname, GLint * data) nogil
cdef glGetIntegerv_type glGetIntegerv

ctypedef void (__stdcall *glGetProgramInfoLog_type)(GLuint  program, GLsizei  bufSize, GLsizei * length, GLchar * infoLog) nogil
cdef glGetProgramInfoLog_type glGetProgramInfoLog

ctypedef void (__stdcall *glGetProgramiv_type)(GLuint  program, GLenum  pname, GLint * params) nogil
cdef glGetProgramiv_type glGetProgramiv

ctypedef void (__stdcall *glGetQueryObjectuiv_type)(GLuint  id, GLenum  pname, GLuint * params) nogil
cdef glGetQueryObjectuiv_type glGetQueryObjectuiv

ctypedef void (__stdcall *glGetQueryiv_type)(GLenum  target, GLenum  pname, GLint * params) nogil
cdef glGetQueryiv_type glGetQueryiv

ctypedef void (__stdcall *glGetRenderbufferParameteriv_type)(GLenum  target, GLenum  pname, GLint * params) nogil
cdef glGetRenderbufferParameteriv_type glGetRenderbufferParameteriv

ctypedef void (__stdcall *glGetShaderInfoLog_type)(GLuint  shader, GLsizei  bufSize, GLsizei * length, GLchar * infoLog) nogil
cdef glGetShaderInfoLog_type glGetShaderInfoLog

ctypedef void (__stdcall *glGetShaderSource_type)(GLuint  shader, GLsizei  bufSize, GLsizei * length, GLchar * source) nogil
cdef glGetShaderSource_type glGetShaderSource

ctypedef void (__stdcall *glGetShaderiv_type)(GLuint  shader, GLenum  pname, GLint * params) nogil
cdef glGetShaderiv_type glGetShaderiv

ctypedef const GLubyte * (__stdcall *glGetString_type)(GLenum  name) nogil
cdef glGetString_type glGetString

ctypedef const GLubyte * (__stdcall *glGetStringi_type)(GLenum  name, GLuint  index) nogil
cdef glGetStringi_type glGetStringi

ctypedef void (__stdcall *glGetTexParameterfv_type)(GLenum  target, GLenum  pname, GLfloat * params) nogil
cdef glGetTexParameterfv_type glGetTexParameterfv

ctypedef void (__stdcall *glGetTexParameteriv_type)(GLenum  target, GLenum  pname, GLint * params) nogil
cdef glGetTexParameteriv_type glGetTexParameteriv

ctypedef void (__stdcall *glGetTransformFeedbackVarying_type)(GLuint  program, GLuint  index, GLsizei  bufSize, GLsizei * length, GLsizei * size, GLenum * type, GLchar * name) nogil
cdef glGetTransformFeedbackVarying_type glGetTransformFeedbackVarying

ctypedef GLint (__stdcall *glGetUniformLocation_type)(GLuint  program, const GLchar * name) nogil
cdef glGetUniformLocation_type glGetUniformLocation

ctypedef void (__stdcall *glGetUniformfv_type)(GLuint  program, GLint  location, GLfloat * params) nogil
cdef glGetUniformfv_type glGetUniformfv

ctypedef void (__stdcall *glGetUniformiv_type)(GLuint  program, GLint  location, GLint * params) nogil
cdef glGetUniformiv_type glGetUniformiv

ctypedef void (__stdcall *glGetUniformuiv_type)(GLuint  program, GLint  location, GLuint * params) nogil
cdef glGetUniformuiv_type glGetUniformuiv

ctypedef void (__stdcall *glGetVertexAttribIiv_type)(GLuint  index, GLenum  pname, GLint * params) nogil
cdef glGetVertexAttribIiv_type glGetVertexAttribIiv

ctypedef void (__stdcall *glGetVertexAttribIuiv_type)(GLuint  index, GLenum  pname, GLuint * params) nogil
cdef glGetVertexAttribIuiv_type glGetVertexAttribIuiv

ctypedef void (__stdcall *glGetVertexAttribPointerv_type)(GLuint  index, GLenum  pname, void ** pointer) nogil
cdef glGetVertexAttribPointerv_type glGetVertexAttribPointerv

ctypedef void (__stdcall *glGetVertexAttribfv_type)(GLuint  index, GLenum  pname, GLfloat * params) nogil
cdef glGetVertexAttribfv_type glGetVertexAttribfv

ctypedef void (__stdcall *glGetVertexAttribiv_type)(GLuint  index, GLenum  pname, GLint * params) nogil
cdef glGetVertexAttribiv_type glGetVertexAttribiv

ctypedef void (__stdcall *glHint_type)(GLenum  target, GLenum  mode) nogil
cdef glHint_type glHint

ctypedef GLboolean (__stdcall *glIsBuffer_type)(GLuint  buffer) nogil
cdef glIsBuffer_type glIsBuffer

ctypedef GLboolean (__stdcall *glIsEnabled_type)(GLenum  cap) nogil
cdef glIsEnabled_type glIsEnabled

ctypedef GLboolean (__stdcall *glIsFramebuffer_type)(GLuint  framebuffer) nogil
cdef glIsFramebuffer_type glIsFramebuffer

ctypedef GLboolean (__stdcall *glIsProgram_type)(GLuint  program) nogil
cdef glIsProgram_type glIsProgram

ctypedef GLboolean (__stdcall *glIsQuery_type)(GLuint  id) nogil
cdef glIsQuery_type glIsQuery

ctypedef GLboolean (__stdcall *glIsRenderbuffer_type)(GLuint  renderbuffer) nogil
cdef glIsRenderbuffer_type glIsRenderbuffer

ctypedef GLboolean (__stdcall *glIsShader_type)(GLuint  shader) nogil
cdef glIsShader_type glIsShader

ctypedef GLboolean (__stdcall *glIsTexture_type)(GLuint  texture) nogil
cdef glIsTexture_type glIsTexture

ctypedef GLboolean (__stdcall *glIsVertexArray_type)(GLuint  array) nogil
cdef glIsVertexArray_type glIsVertexArray

ctypedef void (__stdcall *glLineWidth_type)(GLfloat  width) nogil
cdef glLineWidth_type glLineWidth

ctypedef void (__stdcall *glLinkProgram_type)(GLuint  program) nogil
cdef glLinkProgram_type glLinkProgram

ctypedef void * (__stdcall *glMapBufferRange_type)(GLenum  target, GLintptr  offset, GLsizeiptr  length, GLbitfield  access) nogil
cdef glMapBufferRange_type glMapBufferRange

ctypedef void (__stdcall *glPixelStorei_type)(GLenum  pname, GLint  param) nogil
cdef glPixelStorei_type glPixelStorei

ctypedef void (__stdcall *glPolygonOffset_type)(GLfloat  factor, GLfloat  units) nogil
cdef glPolygonOffset_type glPolygonOffset

ctypedef void (__stdcall *glReadBuffer_type)(GLenum  src) nogil
cdef glReadBuffer_type glReadBuffer

ctypedef void (__stdcall *glReadPixels_type)(GLint  x, GLint  y, GLsizei  width, GLsizei  height, GLenum  format, GLenum  type, void * pixels) nogil
cdef glReadPixels_type glReadPixels

ctypedef void (__stdcall *glRenderbufferStorage_type)(GLenum  target, GLenum  internalformat, GLsizei  width, GLsizei  height) nogil
cdef glRenderbufferStorage_type glRenderbufferStorage

ctypedef void (__stdcall *glRenderbufferStorageMultisample_type)(GLenum  target, GLsizei  samples, GLenum  internalformat, GLsizei  width, GLsizei  height) nogil
cdef glRenderbufferStorageMultisample_type glRenderbufferStorageMultisample

ctypedef void (__stdcall *glSampleCoverage_type)(GLfloat  value, GLboolean  invert) nogil
cdef glSampleCoverage_type glSampleCoverage

ctypedef void (__stdcall *glScissor_type)(GLint  x, GLint  y, GLsizei  width, GLsizei  height) nogil
cdef glScissor_type glScissor

ctypedef void (__stdcall *glShaderSource_type)(GLuint  shader, GLsizei  count, const GLchar *const* string, const GLint * length) nogil
cdef glShaderSource_type glShaderSource

ctypedef void (__stdcall *glStencilFunc_type)(GLenum  func, GLint  ref, GLuint  mask) nogil
cdef glStencilFunc_type glStencilFunc

ctypedef void (__stdcall *glStencilFuncSeparate_type)(GLenum  face, GLenum  func, GLint  ref, GLuint  mask) nogil
cdef glStencilFuncSeparate_type glStencilFuncSeparate

ctypedef void (__stdcall *glStencilMask_type)(GLuint  mask) nogil
cdef glStencilMask_type glStencilMask

ctypedef void (__stdcall *glStencilMaskSeparate_type)(GLenum  face, GLuint  mask) nogil
cdef glStencilMaskSeparate_type glStencilMaskSeparate

ctypedef void (__stdcall *glStencilOp_type)(GLenum  fail, GLenum  zfail, GLenum  zpass) nogil
cdef glStencilOp_type glStencilOp

ctypedef void (__stdcall *glStencilOpSeparate_type)(GLenum  face, GLenum  sfail, GLenum  dpfail, GLenum  dppass) nogil
cdef glStencilOpSeparate_type glStencilOpSeparate

ctypedef void (__stdcall *glTexImage2D_type)(GLenum  target, GLint  level, GLint  internalformat, GLsizei  width, GLsizei  height, GLint  border, GLenum  format, GLenum  type, const void * pixels) nogil
cdef glTexImage2D_type glTexImage2D

ctypedef void (__stdcall *glTexImage3D_type)(GLenum  target, GLint  level, GLint  internalformat, GLsizei  width, GLsizei  height, GLsizei  depth, GLint  border, GLenum  format, GLenum  type, const void * pixels) nogil
cdef glTexImage3D_type glTexImage3D

ctypedef void (__stdcall *glTexParameterf_type)(GLenum  target, GLenum  pname, GLfloat  param) nogil
cdef glTexParameterf_type glTexParameterf

ctypedef void (__stdcall *glTexParameterfv_type)(GLenum  target, GLenum  pname, const GLfloat * params) nogil
cdef glTexParameterfv_type glTexParameterfv

ctypedef void (__stdcall *glTexParameteri_type)(GLenum  target, GLenum  pname, GLint  param) nogil
cdef glTexParameteri_type glTexParameteri

ctypedef void (__stdcall *glTexParameteriv_type)(GLenum  target, GLenum  pname, const GLint * params) nogil
cdef glTexParameteriv_type glTexParameteriv

ctypedef void (__stdcall *glTexSubImage2D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLsizei  width, GLsizei  height, GLenum  format, GLenum  type, const void * pixels) nogil
cdef glTexSubImage2D_type glTexSubImage2D

ctypedef void (__stdcall *glTexSubImage3D_type)(GLenum  target, GLint  level, GLint  xoffset, GLint  yoffset, GLint  zoffset, GLsizei  width, GLsizei  height, GLsizei  depth, GLenum  format, GLenum  type, const void * pixels) nogil
cdef glTexSubImage3D_type glTexSubImage3D

ctypedef void (__stdcall *glTransformFeedbackVaryings_type)(GLuint  program, GLsizei  count, const GLchar *const* varyings, GLenum  bufferMode) nogil
cdef glTransformFeedbackVaryings_type glTransformFeedbackVaryings

ctypedef void (__stdcall *glUniform1f_type)(GLint  location, GLfloat  v0) nogil
cdef glUniform1f_type glUniform1f

ctypedef void (__stdcall *glUniform1fv_type)(GLint  location, GLsizei  count, const GLfloat * value) nogil
cdef glUniform1fv_type glUniform1fv

ctypedef void (__stdcall *glUniform1i_type)(GLint  location, GLint  v0) nogil
cdef glUniform1i_type glUniform1i

ctypedef void (__stdcall *glUniform1iv_type)(GLint  location, GLsizei  count, const GLint * value) nogil
cdef glUniform1iv_type glUniform1iv

ctypedef void (__stdcall *glUniform1ui_type)(GLint  location, GLuint  v0) nogil
cdef glUniform1ui_type glUniform1ui

ctypedef void (__stdcall *glUniform1uiv_type)(GLint  location, GLsizei  count, const GLuint * value) nogil
cdef glUniform1uiv_type glUniform1uiv

ctypedef void (__stdcall *glUniform2f_type)(GLint  location, GLfloat  v0, GLfloat  v1) nogil
cdef glUniform2f_type glUniform2f

ctypedef void (__stdcall *glUniform2fv_type)(GLint  location, GLsizei  count, const GLfloat * value) nogil
cdef glUniform2fv_type glUniform2fv

ctypedef void (__stdcall *glUniform2i_type)(GLint  location, GLint  v0, GLint  v1) nogil
cdef glUniform2i_type glUniform2i

ctypedef void (__stdcall *glUniform2iv_type)(GLint  location, GLsizei  count, const GLint * value) nogil
cdef glUniform2iv_type glUniform2iv

ctypedef void (__stdcall *glUniform2ui_type)(GLint  location, GLuint  v0, GLuint  v1) nogil
cdef glUniform2ui_type glUniform2ui

ctypedef void (__stdcall *glUniform2uiv_type)(GLint  location, GLsizei  count, const GLuint * value) nogil
cdef glUniform2uiv_type glUniform2uiv

ctypedef void (__stdcall *glUniform3f_type)(GLint  location, GLfloat  v0, GLfloat  v1, GLfloat  v2) nogil
cdef glUniform3f_type glUniform3f

ctypedef void (__stdcall *glUniform3fv_type)(GLint  location, GLsizei  count, const GLfloat * value) nogil
cdef glUniform3fv_type glUniform3fv

ctypedef void (__stdcall *glUniform3i_type)(GLint  location, GLint  v0, GLint  v1, GLint  v2) nogil
cdef glUniform3i_type glUniform3i

ctypedef void (__stdcall *glUniform3iv_type)(GLint  location, GLsizei  count, const GLint * value) nogil
cdef glUniform3iv_type glUniform3iv

ctypedef void (__stdcall *glUniform3ui_type)(GLint  location, GLuint  v0, GLuint  v1, GLuint  v2) nogil
cdef glUniform3ui_type glUniform3ui

ctypedef void (__stdcall *glUniform3uiv_type)(GLint  location, GLsizei  count, const GLuint * value) nogil
cdef glUniform3uiv_type glUniform3uiv

ctypedef void (__stdcall *glUniform4f_type)(GLint  location, GLfloat  v0, GLfloat  v1, GLfloat  v2, GLfloat  v3) nogil
cdef glUniform4f_type glUniform4f

ctypedef void (__stdcall *glUniform4fv_type)(GLint  location, GLsizei  count, const GLfloat * value) nogil
cdef glUniform4fv_type glUniform4fv

ctypedef void (__stdcall *glUniform4i_type)(GLint  location, GLint  v0, GLint  v1, GLint  v2, GLint  v3) nogil
cdef glUniform4i_type glUniform4i

ctypedef void (__stdcall *glUniform4iv_type)(GLint  location, GLsizei  count, const GLint * value) nogil
cdef glUniform4iv_type glUniform4iv

ctypedef void (__stdcall *glUniform4ui_type)(GLint  location, GLuint  v0, GLuint  v1, GLuint  v2, GLuint  v3) nogil
cdef glUniform4ui_type glUniform4ui

ctypedef void (__stdcall *glUniform4uiv_type)(GLint  location, GLsizei  count, const GLuint * value) nogil
cdef glUniform4uiv_type glUniform4uiv

ctypedef void (__stdcall *glUniformMatrix2fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix2fv_type glUniformMatrix2fv

ctypedef void (__stdcall *glUniformMatrix2x3fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix2x3fv_type glUniformMatrix2x3fv

ctypedef void (__stdcall *glUniformMatrix2x4fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix2x4fv_type glUniformMatrix2x4fv

ctypedef void (__stdcall *glUniformMatrix3fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix3fv_type glUniformMatrix3fv

ctypedef void (__stdcall *glUniformMatrix3x2fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix3x2fv_type glUniformMatrix3x2fv

ctypedef void (__stdcall *glUniformMatrix3x4fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix3x4fv_type glUniformMatrix3x4fv

ctypedef void (__stdcall *glUniformMatrix4fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix4fv_type glUniformMatrix4fv

ctypedef void (__stdcall *glUniformMatrix4x2fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix4x2fv_type glUniformMatrix4x2fv

ctypedef void (__stdcall *glUniformMatrix4x3fv_type)(GLint  location, GLsizei  count, GLboolean  transpose, const GLfloat * value) nogil
cdef glUniformMatrix4x3fv_type glUniformMatrix4x3fv

ctypedef GLboolean (__stdcall *glUnmapBuffer_type)(GLenum  target) nogil
cdef glUnmapBuffer_type glUnmapBuffer

ctypedef void (__stdcall *glUseProgram_type)(GLuint  program) nogil
cdef glUseProgram_type glUseProgram

ctypedef void (__stdcall *glValidateProgram_type)(GLuint  program) nogil
cdef glValidateProgram_type glValidateProgram

ctypedef void (__stdcall *glVertexAttrib1f_type)(GLuint  index, GLfloat  x) nogil
cdef glVertexAttrib1f_type glVertexAttrib1f

ctypedef void (__stdcall *glVertexAttrib1fv_type)(GLuint  index, const GLfloat * v) nogil
cdef glVertexAttrib1fv_type glVertexAttrib1fv

ctypedef void (__stdcall *glVertexAttrib2f_type)(GLuint  index, GLfloat  x, GLfloat  y) nogil
cdef glVertexAttrib2f_type glVertexAttrib2f

ctypedef void (__stdcall *glVertexAttrib2fv_type)(GLuint  index, const GLfloat * v) nogil
cdef glVertexAttrib2fv_type glVertexAttrib2fv

ctypedef void (__stdcall *glVertexAttrib3f_type)(GLuint  index, GLfloat  x, GLfloat  y, GLfloat  z) nogil
cdef glVertexAttrib3f_type glVertexAttrib3f

ctypedef void (__stdcall *glVertexAttrib3fv_type)(GLuint  index, const GLfloat * v) nogil
cdef glVertexAttrib3fv_type glVertexAttrib3fv

ctypedef void (__stdcall *glVertexAttrib4f_type)(GLuint  index, GLfloat  x, GLfloat  y, GLfloat  z, GLfloat  w) nogil
cdef glVertexAttrib4f_type glVertexAttrib4f

ctypedef void (__stdcall *glVertexAttrib4fv_type)(GLuint  index, const GLfloat * v) nogil
cdef glVertexAttrib4fv_type glVertexAttrib4fv

ctypedef void (__stdcall *glVertexAttribI4i_type)(GLuint  index, GLint  x, GLint  y, GLint  z, GLint  w) nogil
cdef glVertexAttribI4i_type glVertexAttribI4i

ctypedef void (__stdcall *glVertexAttribI4iv_type)(GLuint  index, const GLint * v) nogil
cdef glVertexAttribI4iv_type glVertexAttribI4iv

ctypedef void (__stdcall *glVertexAttribI4ui_type)(GLuint  index, GLuint  x, GLuint  y, GLuint  z, GLuint  w) nogil
cdef glVertexAttribI4ui_type glVertexAttribI4ui

ctypedef void (__stdcall *glVertexAttribI4uiv_type)(GLuint  index, const GLuint * v) nogil
cdef glVertexAttribI4uiv_type glVertexAttribI4uiv

ctypedef void (__stdcall *glVertexAttribIPointer_type)(GLuint  index, GLint  size, GLenum  type, GLsizei  stride, const void * pointer) nogil
cdef glVertexAttribIPointer_type glVertexAttribIPointer

ctypedef void (__stdcall *glVertexAttribPointer_type)(GLuint  index, GLint  size, GLenum  type, GLboolean  normalized, GLsizei  stride, const void * pointer) nogil
cdef glVertexAttribPointer_type glVertexAttribPointer

ctypedef void (__stdcall *glViewport_type)(GLint  x, GLint  y, GLsizei  width, GLsizei  height) nogil
cdef glViewport_type glViewport



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# This file contains code that handles the execution of python code
# contained within the script file. It also handles rolling back the
# game state to some time in the past.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *

from typing import Optional

import marshal
import random
import weakref
import re
import sys
import time
import io
import types
import copyreg
import functools

import renpy

# Deleted is a singleton object that's used to represent an object that has
# been deleted from the store.

class StoreDeleted(object):

    def __reduce__(self):
        if PY2:
            return b"deleted"
        else:
            return "deleted"


deleted = StoreDeleted()


# Code that computes reachable objects, which is used to filter
# the rollback list before rollback or serialization.


class SlottedNoRollback(object):
    """
    :doc: norollback class

    Instances of classes inheriting from this class do not participate
    in rollback. The difference between this and :class:`NoRollback` is that
    this class does not have an associated dictionary, hence can be used
    with ``__slots__`` to reduce memory usage.

    Objects reachable through an instance of a NoRollback class only participate
    in rollback if they are reachable through other paths.
    """
    __slots__ = ()


class NoRollback(SlottedNoRollback):
    """
    :doc: norollback class

    Instances of this class, and classes inheriting from this class,
    do not participate in rollback. Objects reachable through an instance
    of a NoRollback class only participate in rollback if they are
    reachable through other paths.
    """

# parents = [ ]

class AlwaysRollback(renpy.revertable.RevertableObject):
    """
    This is a revertible object that always participates in rollback.
    It's used when a revertable object is created by an object that
    doesn't participate in the rollback system.
    """

    def __new__(cls, *args, **kwargs):
        self = super(AlwaysRollback, cls).__new__(cls)

        log = renpy.game.log
        if log is not None:
            del log.mutated[id(self)]

        return self


NOROLLBACK_TYPES = tuple() # type: tuple[type, type, type, type, type]


def reached(obj, reachable, wait):
    """
    @param obj: The object that was reached.

    `reachable`
        A map from id(obj) to int. The int is 1 if the object was reached
        normally, and 0 if it was reached, but inherits from NoRollback.
    """

    if wait:
        wait()

    idobj = id(obj)

    if idobj in reachable:
        return

    reachable[idobj] = obj

    if isinstance(obj, NOROLLBACK_TYPES):
        return

    try:
        nosave = getattr(obj, "nosave", None)

        if nosave is not None:
            for k, v in vars(obj).items():
                if k not in nosave:
                    reached(v, reachable, wait)

        else:

            # Fields have to be indexed by strings, so no need to check if
            # the filed is reached.
            for v in vars(obj).values():
                reached(v, reachable, wait)

    except Exception:
        pass

    # Below this, we only consider containers with a defined size.
    try:
        if (not len(obj)) or isinstance(obj, basestring):
            return
    except Exception:
        return

    try:
        # Treat as iterable.
        for v in obj.__iter__():
            reached(v, reachable, wait)
    except Exception:
        pass

    try:
        # Treat as dict.
        for v in obj.values():
            reached(v, reachable, wait)
    except Exception:
        pass


def reached_vars(store, reachable, wait):
    """
    Marks everything reachable from the variables in the store
    or from the context info objects as reachable.

    `store`
        A map from variable name to variable value.

    `reachable`
        A dictionary that will be filled in with a map from id(obj) to obj.
    """

    for v in store.values():
        reached(v, reachable, wait)

    for c in renpy.game.contexts:
        reached(c.info, reachable, wait)
        reached(c.music, reachable, wait)
        for d in c.dynamic_stack:
            for v in d.values():
                reached(v, reachable, wait)

# This is the code that actually handles the logging and managing
# of the rollbacks.

generation = time.time()
serial = 0


rng = renpy.revertable.DetRandom()

class Rollback(renpy.object.Object):
    """
    Allows the state of the game to be rolled back to the point just
    before a node began executing.

    @ivar context: A shallow copy of the context we were in before
    we started executing the node. (Shallow copy also includes
    a copy of the associated SceneList.)

    @ivar objects: A list of tuples, each containing an object and a
    token of information that, when passed to the rollback method on
    that object, causes that object to rollback.

    @ivar store: A list of updates to store that will cause the state
    of the store to be rolled back to the start of node
    execution. This is a list of tuples, either (key, value) tuples
    representing a value that needs to be assigned to a key, or (key,)
    tuples that mean the key should be deleted.

    @ivar checkpoint: True if this is a user-visible checkpoint,
    false otherwise.

    @ivar purged: True if purge_unreachable has already been called on
    this Rollback, False otherwise.

    @ivar random: A list of random numbers that were generated during the
    execution of this element.
    """

    __version__ = 5

    identifier = None
    not_greedy = False

    def __init__(self):

        super(Rollback, self).__init__()

        self.context = renpy.game.context().rollback_copy()

        self.objects = [ ]
        self.purged = False
        self.random = [ ]
        self.forward = None

        # A map of maps name -> (variable -> value)
        self.stores = { }

        # A map from store name to the changes to ever_been_changed that
        # need to be reverted.
        self.delta_ebc = { }

        # If true, we retain the data in this rollback when a load occurs.
        self.retain_after_load = False

        # True if this is a checkpoint we can roll back to.
        self.checkpoint = False

        # True if this is a hard checkpoint, where the rollback counter
        # decreases.
        self.hard_checkpoint = False

        # True if this is a not-greedy checkpoint, which should end
        # rollbacks that occur in greedy mode.
        self.not_greedy = False

        # A unique identifier for this rollback object.

        global serial
        self.identifier = (generation, serial)
        serial += 1

    def after_upgrade(self, version):

        if version < 2:
            self.stores = { "store" : { } }

            for i in self.store: # type: ignore
                if len(i) == 2:
                    k, v = i
                    self.stores["store"][k] = v
                else:
                    k, = i
                    self.stores["store"][k] = deleted

        if version < 3:
            self.retain_after_load = False

        if version < 4:
            self.hard_checkpoint = self.checkpoint

        if version < 5:
            self.delta_ebc = { }

    def purge_unreachable(self, reachable, wait):
        """
        Adds objects that are reachable from the store of this
        rollback to the set of reachable objects, and purges
        information that is stored about totally unreachable objects.

        Returns True if this is the first time this method has been
        called, or False if it has already been called once before.
        """

        if self.purged:
            return False

        self.purged = True

        # Add objects reachable from the stores. (Objects that might be
        # unreachable at the moment.)
        for changes in self.stores.values():
            for _k, v in changes.items():
                if v is not deleted:
                    reached(v, reachable, wait)

        # Add in objects reachable through the context.
        reached(self.context.info, reachable, wait)
        for d in self.context.dynamic_stack:
            for v in d.values():
                reached(v, reachable, wait)

        # Add in objects reachable through displayables.
        reached(self.context.scene_lists.get_all_displayables(), reachable, wait)

        # Purge object update information for unreachable objects.
        new_objects = [ ]

        objects_changed = True
        seen = set()

        while objects_changed:

            objects_changed = False

            for o, rb in self.objects:

                id_o = id(o)

                if id_o in seen or id_o not in reachable:
                    continue

                seen.add(id_o)

                if isinstance(o, NOROLLBACK_TYPES):
                    continue

                objects_changed = True

                new_objects.append((o, rb))
                reached(rb, reachable, wait)

        del self.objects[:]
        self.objects.extend(new_objects)

        return True

    def rollback(self):
        """
        Reverts the state of the game to what it was at the start of the
        previous checkpoint.
        """

        store_dicts = renpy.python.store_dicts

        for obj, roll in reversed(self.objects):

            if roll is not None:
                obj._rollback(roll)

        for name, changes in self.stores.items():
            store = store_dicts.get(name, None)
            if store is None:
                continue

            for name, value in changes.items():
                if value is deleted:
                    if name in store:
                        del store[name]
                else:
                    store[name] = value

        for name, changes in self.delta_ebc.items():

            store = store_dicts.get(name, None)
            if store is None:
                continue

            store.ever_been_changed -= changes

        rng.pushback(self.random)

        renpy.game.contexts.pop()
        renpy.game.contexts.append(self.context)

    def rollback_control(self):
        """
        This rolls back only the control information, while leaving
        the data information intact.
        """

        renpy.game.contexts.pop()
        renpy.game.contexts.append(self.context)


class RollbackLog(renpy.object.Object):
    """
    This class manages the list of Rollback objects.

    @ivar log: The log of rollback objects.

    @ivar current: The current rollback object. (Equivalent to
    log[-1])

    @ivar rollback_limit: The number of steps left that we can
    interactively rollback.

    Not serialized:

    @ivar mutated: A dictionary that maps object ids to a tuple of
    (weakref to object, information needed to rollback that object)
    """

    __version__ = 5

    nosave = [ 'old_store', 'mutated', 'identifier_cache' ]
    identifier_cache = None
    force_checkpoint = False

    def __init__(self):

        super(RollbackLog, self).__init__()

        self.log = [ ]
        self.current = None
        self.mutated = { }
        self.rollback_limit = 0
        self.rollback_is_fixed = False
        self.checkpointing_suspended = False
        self.fixed_rollback_boundary = None
        self.forward = [ ]
        self.old_store = { }

        # Did we just do a roll forward?
        self.rolled_forward = False

        # Reset the RNG on the creation of a new game.
        rng.reset()

        # True if we should retain data from here to the next checkpoint
        # on load.
        self.retain_after_load_flag = False

        # Has there been an interaction since the last time this log was
        # reset?
        self.did_interaction = True

        # Should we force a checkpoint before completing the current
        # statement.
        self.force_checkpoint = False

    def after_setstate(self):
        self.mutated = { }
        self.rolled_forward = False

    def after_upgrade(self, version):
        if version < 2:
            self.ever_been_changed = { "store" : set(self.ever_been_changed) }
        if version < 3:
            self.rollback_is_fixed = False
            self.fixed_rollback_boundary = None
        if version < 4:
            self.retain_after_load_flag = False

        if version < 5:

            # We changed what the rollback limit represents, so recompute it
            # here.
            if self.rollback_limit:
                nrbl = 0

                for rb in self.log[-self.rollback_limit:]:
                    if rb.hard_checkpoint:
                        nrbl += 1

                self.rollback_limit = nrbl

    def begin(self, force=False):
        """
        Called before a node begins executing, to indicate that the
        state needs to be saved for rollbacking.
        """

        self.identifier_cache = None

        context = renpy.game.context()

        if not context.rollback:
            return

        # We only begin a checkpoint if the previous statement reached a checkpoint,
        # or an interaction took place. (Or we're forced.)
        ignore = True

        if force:
            ignore = False
        elif self.did_interaction:
            ignore = False
        elif self.current is not None:
            if self.current.checkpoint:
                ignore = False
            elif self.current.retain_after_load:
                ignore = False

        if ignore:
            return

        self.did_interaction = False

        if self.current is not None:
            self.complete(True)
        else:
            renpy.python.begin_stores()

        # If the log is too long, prune it.
        while len(self.log) > renpy.config.rollback_length:
            self.log.pop(0)

        # check for the end of fixed rollback
        if len(self.log) >= 2:
            if self.log[-2].context.current == self.fixed_rollback_boundary:
                self.rollback_is_fixed = False

        # A lack of rollback data in fixed rollback mode ends rollback.
        if self.rollback_is_fixed and not self.forward:
            self.fixed_rollback_boundary = self.current.context.current
            self.rollback_is_fixed = False

        self.current = Rollback()
        self.current.retain_after_load = self.retain_after_load_flag

        self.log.append(self.current)

        self.mutated.clear()

        # Flag a mutation as having happened. This is used by the
        # save code.
        renpy.revertable.mutate_flag = True

        self.rolled_forward = False

    def replace_node(self, old, new):
        """
        Replaces references to the `old` ast node with a reference to the
        `new` ast node.
        """

        for i in self.log:
            i.context.replace_node(old, new)

    def complete(self, begin=False):
        """
        Called after a node is finished executing, before a save
        begins, or right before a rollback is attempted. This may be
        called more than once between calls to begin, and should always
        be called after an update to the store but before a rollback
        occurs.

        `begin`
            Should be true if called from begin().
        """

        if self.force_checkpoint:
            self.checkpoint(hard=False)
            self.force_checkpoint = False

        # Update self.current.stores with the changes from each store.
        # Also updates .ever_been_changed.
        for name, sd in renpy.python.store_dicts.items():
            delta = sd.get_changes(begin)
            if delta:
                self.current.stores[name], self.current.delta_ebc[name] = delta

        # Update the list of mutated objects and what we need to do to
        # restore them.

        for _i in range(4):

            del self.current.objects[:]

            try:
                for _k, v in self.mutated.items():

                    if v is None:
                        continue

                    (ref, clean) = v

                    obj = ref()
                    if obj is None:
                        continue

                    compressed = obj._compress(clean)
                    self.current.objects.append((obj, compressed))

                break

            except RuntimeError:
                # This can occur when self.mutated is changed as we're
                # iterating over it.
                pass

    def get_roots(self):
        """
        Return a map giving the current roots of the store. This is a
        map from a variable name in the store to the value of that
        variable. A variable is only in this map if it has ever been
        changed since the init phase finished.
        """

        rv = { }

        for store_name, sd in renpy.python.store_dicts.items():
            for name in sd.ever_been_changed:
                if name in sd:
                    rv[store_name + "." + name] = sd[name]
                else:
                    rv[store_name + "." + name] = deleted

        for i in reversed(renpy.game.contexts[1:]):
            i.pop_dynamic_roots(rv)

        return rv

    def purge_unreachable(self, roots, wait=None):
        """
        This is called to purge objects that are unreachable from the
        roots from the object rollback lists inside the Rollback entries.

        This should be called immediately after complete(), so that there
        are no changes queued up.
        """

        # This needs to be set late, so that StoreModule is available.

        global NOROLLBACK_TYPES
        NOROLLBACK_TYPES = (types.ModuleType, renpy.python.StoreModule, SlottedNoRollback, io.IOBase, type)

        reachable = { }

        reached_vars(roots, reachable, wait)

        revlog = self.log[:]
        revlog.reverse()

        for i in revlog:
            if not i.purge_unreachable(reachable, wait):
                break

        # Break any cycles.
        reachable.clear()

    def in_rollback(self):
        if self.forward:
            return True
        else:
            return False

    def in_fixed_rollback(self):
        return self.rollback_is_fixed

    def forward_info(self):
        """
        Returns the current forward info, if any.
        """

        if self.forward:

            name, data = self.forward[0]

            if self.current.context.current == name:
                return data

        return None

    def checkpoint(self, data=None, keep_rollback=False, hard=True):
        """
        Called to indicate that this is a checkpoint, which means
        that the user may want to rollback to just before this
        node.
        """

        if self.checkpointing_suspended:
            hard = False

        if hard:
            self.retain_after_load_flag = False

        if self.current.checkpoint:
            return

        if not renpy.game.context().rollback:
            return

        self.current.checkpoint = True

        if hard and (not self.current.hard_checkpoint):
            if self.rollback_limit < renpy.config.hard_rollback_limit:
                self.rollback_limit += 1

            if hard == "not_greedy":
                self.current.not_greedy = True
            else:
                self.current.hard_checkpoint = hard

        if self.in_fixed_rollback() and self.forward:
            # use data from the forward stack
            fwd_name, fwd_data = self.forward[0]
            if self.current.context.current == fwd_name:
                self.current.forward = fwd_data
                self.forward.pop(0)
            else:
                self.current.forward = data
                del self.forward[:]

        elif data is not None:
            if self.forward:
                # If the data is the same, pop it from the forward stack.
                # Otherwise, clear the forward stack.
                fwd_name, fwd_data = self.forward[0]

                if (self.current.context.current == fwd_name
                    and data == fwd_data
                    and (keep_rollback or self.rolled_forward)
                    ):
                    self.forward.pop(0)
                else:
                    del self.forward[:]

            # Log the data in case we roll back again.
            self.current.forward = data

    def suspend_checkpointing(self, flag):
        """
        Called to temporarily suspend checkpointing, so any rollback
        will jump to prior to this statement
        """

        self.checkpointing_suspended = flag

    def block(self, purge=False):
        """
        Called to indicate that the user should not be able to rollback
        through this checkpoint.
        """

        self.rollback_limit = 0
        renpy.game.context().force_checkpoint = True

        if purge:
            del self.log[:]

    def retain_after_load(self):
        """
        Called to return data from this statement until the next checkpoint
        when the game is loaded.
        """

        if renpy.display.predict.predicting:
            return

        self.retain_after_load_flag = True
        self.current.retain_after_load = True

        for rb in reversed(self.log):
            if rb.hard_checkpoint:
                break

            rb.retain_after_load = True

        renpy.game.context().force_checkpoint = True

    def fix_rollback(self):
        if not self.rollback_is_fixed and len(self.log) > 1:
            self.fixed_rollback_boundary = self.log[-2].context.current
        renpy.game.context().force_checkpoint = True

    def can_rollback(self):
        """
        Returns True if we can rollback.
        """

        return self.rollback_limit > 0

    def load_failed(self):
        """
        This is called to try to recover when rollback fails.
        """

        lfl = renpy.config.load_failed_label
        if callable(lfl):
            lfl = lfl()

        if not lfl:
            raise Exception("Couldn't find a place to stop rolling back. Perhaps the script changed in an incompatible way?")

        rb = self.log.pop()
        rb.rollback()

        while renpy.exports.call_stack_depth():
            renpy.exports.pop_call()

        renpy.game.contexts[0].force_checkpoint = True
        renpy.game.contexts[0].goto_label(lfl)

        raise renpy.game.RestartTopContext()

    def rollback(self, checkpoints, force=False, label=None, greedy=True, on_load=False, abnormal=True, current_label=None):
        """
        This rolls the system back to the first valid rollback point
        after having rolled back past the specified number of checkpoints.

        If we're currently executing code, it's expected that complete()
        will be called before a rollback is attempted.

        force makes us throw an exception if we can't find a place to stop
        rolling back, otherwise if we run out of log this call has no
        effect.

        `label`
            A label that is called after rollback has finished, if the
            label exists.

        `greedy`
            If true, rollback will keep going until just after the last
            checkpoint. If False, it will stop immediately before the
            current statement.

        `on_load`
            Should be true if this rollback is being called in response to a
            load. Used to implement .retain_after_load()

        `abnormal`
            If true, treats this as an abnormal event, suppressing transitions
            and so on.

        `current_label`
            A lable that is called when control returns to the current statement,
            after rollback. (At most one of `current_label` and `label` can be
            provided.)
        """

        # If we have exceeded the rollback limit, and don't have force,
        # give up.
        if checkpoints and (self.rollback_limit <= 0) and (not force):
            return

        self.suspend_checkpointing(False)
        # will always rollback to before suspension

        self.purge_unreachable(self.get_roots())

        revlog = [ ]

        # Find the place to roll back to.
        while self.log:
            rb = self.log.pop()
            revlog.append(rb)

            if rb.hard_checkpoint:
                self.rollback_limit -= 1

            if rb.hard_checkpoint or (on_load and rb.checkpoint):
                checkpoints -= 1

            if checkpoints <= 0:
                if renpy.game.script.has_label(rb.context.current):
                    break

        else:
            # Otherwise, just give up.

            revlog.reverse()
            self.log.extend(revlog)

            if force:
                self.load_failed()
            else:
                print("Can't find a place to rollback to. Not rolling back.")

            return

        force_checkpoint = False

        # Try to rollback to just after the previous checkpoint.
        while greedy and self.log and (self.rollback_limit > 0):

            rb = self.log[-1]

            if not renpy.game.script.has_label(rb.context.current):
                break

            if rb.hard_checkpoint:
                break

            if rb.not_greedy:
                break

            revlog.append(self.log.pop())

        # Decide if we're replacing the current context (rollback command),
        # or creating a new set of contexts (loading).
        if renpy.game.context().rollback:
            replace_context = False
            other_contexts = [ ]

        else:
            replace_context = True
            other_contexts = renpy.game.contexts[1:]
            renpy.game.contexts = renpy.game.contexts[0:1]

        if on_load and revlog[-1].retain_after_load:
            retained = revlog.pop()
            self.retain_after_load_flag = True
        else:
            retained = None

        come_from = None

        if current_label is not None:
            come_from = renpy.game.context().current
            label = current_label

        # Actually roll things back.
        for rb in revlog:
            rb.rollback()

            if (rb.context.current == self.fixed_rollback_boundary) and (rb.context.current):
                self.rollback_is_fixed = True

            if rb.forward is not None:
                self.forward.insert(0, (rb.context.current, rb.forward))

        if retained is not None:
            retained.rollback_control()
            self.log.append(retained)

        if (label is not None) and (come_from is None):
            come_from = renpy.game.context().current

        if come_from is not None:
            renpy.game.context().come_from(come_from, label)

        # Disable the next transition, as it's pointless. (Only when not used with a label.)
        renpy.game.interface.suppress_transition = abnormal

        # If necessary, reset the RNG.
        if force:
            rng.reset()
            del self.forward[:]

        # Flag that we're in the transition immediately after a rollback.
        renpy.game.after_rollback = abnormal

        # Stop the sounds.
        renpy.audio.audio.rollback()

        # Stop any hiding in progress.
        for i in renpy.game.contexts:
            i.scene_lists.remove_all_hidden()

        renpy.game.contexts.extend(other_contexts)

        renpy.exports.execute_default_statement(False)

        self.mutated.clear()
        renpy.python.begin_stores()

        # Restart the context or the top context.
        if replace_context:

            if force_checkpoint:
                renpy.game.contexts[0].force_checkpoint = True

            self.current = Rollback()
            self.current.context = renpy.game.contexts[0].rollback_copy()

            if self.log is not None:
                self.log.append(self.current)

            raise renpy.game.RestartTopContext()

        else:

            self.current = Rollback()
            self.current.context = renpy.game.context().rollback_copy()

            if self.log is not None:
                self.log.append(self.current)

            if force_checkpoint:
                renpy.game.context().force_checkpoint = True

            raise renpy.game.RestartContext()

    def freeze(self, wait=None):
        """
        This is called to freeze the store and the log, in preparation
        for serialization. The next call on log should either be
        unfreeze (called after a serialization reload) or discard_freeze()
        (called after the save is complete).
        """

        # Purge unreachable objects, so we don't save them.
        self.complete(False)
        roots = self.get_roots()

        self.purge_unreachable(roots, wait=wait)

        # The current is not purged.
        self.current.purged = False

        return roots

    def discard_freeze(self):
        """
        Called to indicate that we will not be restoring from the
        frozen state.
        """

    def unfreeze(self, roots, label=None):
        """
        Used to unfreeze the game state after a load of this log
        object. This call will always throw an exception. If we're
        lucky, it's the one that indicates load was successful.

        @param roots: The roots returned from freeze.

        @param label: The label that is jumped to in the game script
        after rollback has finished, if it exists.
        """

        # Fix up old screens.
        renpy.display.screen.before_restart() # @UndefinedVariable

        # Set us up as the game log.
        renpy.game.log = self

        renpy.python.clean_stores()
        renpy.translation.init_translation()

        store_dicts = renpy.python.store_dicts

        for name, value in roots.items():

            if "." in name:
                store_name, name = name.rsplit(".", 1)
            else:
                store_name = "store"

            if store_name not in store_dicts:
                continue

            store = store_dicts[store_name]
            store.ever_been_changed.add(name)

            if value is deleted:
                if name in store:
                    del store[name]
            else:
                store[name] = value

        # Now, rollback to an acceptable point.

        greedy = renpy.session.pop("_greedy_rollback", False)
        self.rollback(0, force=True, label=label, greedy=greedy, on_load=True)

        # Because of the rollback, we never make it this far.

    def build_identifier_cache(self):

        if self.identifier_cache is not None:
            return

        rollback_limit = self.rollback_limit
        checkpoints = 1

        self.identifier_cache = { }

        for i in reversed(self.log):

            if i.identifier is not None:
                if renpy.game.script.has_label(i.context.current):
                    self.identifier_cache[i.identifier] = checkpoints

            if i.hard_checkpoint:
                checkpoints += 1

            if i.checkpoint:
                rollback_limit -= 1

            if not rollback_limit:
                break

    def get_identifier_checkpoints(self, identifier):
        self.build_identifier_cache()
        return self.identifier_cache.get(identifier, None)



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# http://www.csn.ul.ie/~caolan/publink/winresdump/winresdump/doc/pefile.html
# Contains a reasonable description of the format.

from __future__ import print_function
from renpy.compat import bchr, chr, str, PY2

import struct
import sys
import array
import pefile


# This class performs various operations on memory-loaded binary files,
# including modifications.

class BinFile(object):

    def set_u32(self, addr, value):
        self.a[addr + 0] = (value >> 0) & 0xff
        self.a[addr + 1] = (value >> 8) & 0xff
        self.a[addr + 2] = (value >> 16) & 0xff
        self.a[addr + 3] = (value >> 24) & 0xff

    def u32(self):
        addr = self.addr
        rv = self.a[addr]
        rv |= self.a[addr + 1] << 8
        rv |= self.a[addr + 2] << 16
        rv |= self.a[addr + 3] << 24
        self.addr += 4
        return rv

    def u16(self):
        addr = self.addr
        rv = self.a[addr]
        rv |= self.a[addr + 1] << 8
        self.addr += 2
        return rv

    def u8(self):
        rv = self.a[self.addr]
        self.addr += 1
        return rv

    def name(self):
        c = self.u16()

        rv = u""
        for _i in range(c):
            rv += chr(self.u16())

        return rv

    def seek(self, addr):
        self.addr = addr

    def tostring(self):
        if PY2:
            return self.a.tostring() # type: ignore
        else:
            return self.a.tobytes()

    def substring(self, start, len): # @ReservedAssignment
        if PY2:
            return self.a[start:start + len].tostring() # type: ignore
        else:
            return self.a[start:start + len].tobytes()

    def __init__(self, data):
        self.a = array.array('B')

        if PY2:
            self.a.fromstring(data) # type: ignore
        else:
            self.a.frombytes(data)

##############################################################################
# These functions parse data out of the file. In these functions, offset is
# relative to the start of the file.


# The virtual address of the resource segment.
resource_virtual = 0

# This parses a data block out of the resources.


def parse_data(bf, offset):
    bf.seek(offset)
    data_offset = bf.u32()
    data_len = bf.u32()
    code_page = bf.u32()
    bf.u32()

    l = [ ]

    bf.seek(data_offset - resource_virtual)
    for _i in range(data_len):
        l.append(bchr(bf.u8()))

    return (code_page, b"".join(l))

# This parses a resource directory.


def parse_directory(bf, offset):

    bf.seek(offset)
    char = bf.u32() # @UnusedVariable
    timedate = bf.u32() # @UnusedVariable
    major = bf.u16() # @UnusedVariable
    minor = bf.u16() # @UnusedVariable
    n_named = bf.u16()
    n_id = bf.u16()

    entries = [ ]

    for _i in range(n_named + n_id):
        entries.append((bf.u32(), bf.u32()))

    rv = { }

    for name, value in entries:

        if name & 0x80000000:
            bf.seek((name & 0x7fffffff))
            name = bf.name()

        if value & 0x80000000:
            value = parse_directory(bf, value & 0x7fffffff)
        else:
            value = parse_data(bf, value)

        rv[name] = value

    return rv


##############################################################################
# This utility function displays the tree of resources that have been loaded.
def show_resources(d, prefix):

    if not isinstance(d, dict):
        print(prefix, "Codepage", d[0], "length", len(d[1]))
        return

    for k in d:
        print(prefix, k)
        show_resources(d[k], prefix + "  ")

##############################################################################
# These functions repack the resources into a new resource segment. Here,
# the offset is relative to the start of the resource segment.


class Packer(object):

    def pack(self, d):
        self.data = b""
        self.data_offset = 0

        self.entries = b""
        self.entries_offset = 0

        head = self.pack_dict(d, 0)

        self.data = b""
        self.data_offset = len(head) + len(self.entries)

        self.entries = b""
        self.entries_offset = len(head)

        return self.pack_dict(d, 0) + self.entries + self.data

    def pack_name(self, s):
        rv = self.data_offset + len(self.data)

        l = len(s)
        s = s.encode("utf-16le")
        self.data += struct.pack("<H", l) + s + b"\0\0"

        return rv

    def pack_tuple(self, t):
        codepage, data = t

        rv = len(self.entries) + self.entries_offset

        if len(self.data) % 2:
            self.data += b"P"

        daddr = len(self.data) + self.data_offset

        self.entries += struct.pack("<IIII", daddr + resource_virtual, len(data), codepage, 0)
        self.data += data

        # if len(self.data) % 1 == 1:
        #    self.data += 'P'

        return rv

    def pack_dict(self, d, offset):
        name_entries = sorted((a, b) for a, b in d.items() if isinstance(a, str))
        id_entries = sorted((a, b) for a, b in d.items() if isinstance(a, int))

        rv = struct.pack("<IIHHHH", 0, 0, 4, 0, len(name_entries), len(id_entries))

        offset += len(rv) + (len(name_entries) + len(id_entries)) * 8

        rest = b""

        for (name, value) in name_entries + id_entries:
            if isinstance(name, str):
                name = 0x80000000 | self.pack_name(name)

            if isinstance(value, dict):
                addr = offset | 0x80000000
                packed = self.pack_dict(value, offset)
                offset += len(packed)
                rest += packed
            else:
                addr = self.pack_tuple(value)

            rv += struct.pack("<II", name, addr)

        return rv + rest

##############################################################################
# This loads in an icon file, and returns a dictionary that is suitable for
# use in the resources of an exe file.


def load_icon(fn):
    f = BinFile(open(fn, "rb").read())

    f.seek(0)
    f.u16()
    f.u16()
    count = f.u16()

    rv = { }
    rv[3] = { }

    group = struct.pack("HHH", 0, 1, count)

    for i in range(count):
        width = f.u8()
        height = f.u8()
        colors = f.u8()
        reserved = f.u8()
        planes = f.u16()
        bpp = f.u16()
        size = f.u32()
        offset = f.u32()

        addr = f.addr
        f.seek(offset + 16)
        if not f.u32():
            f.set_u32(offset + 20, 0)

        #rv[3][i + 1] = { 0 : (1252, f.substring(offset, size)) }
        #I do not have justification for the 0x0409 entry ID, other than copying what other files do. Same for codepage of 0
        rv[3][i + 1] = { 1033 : (0, f.substring(offset, size)) }

        group += struct.pack("BBBBHHIH", width, height, colors, reserved,
                             planes, bpp, size, i + 1)

        f.seek(addr)

    #rv[14] = { 1 : { 0 : (1252, group) } }
    rv[14] = { 1 : { 1033 : (0, group) } }

    return rv


##############################################################################
# This is the main function that should be called externally, that copies over
# the icons.
def change_icons(oldexe, icofn):
    global resource_virtual

    pe = pefile.PE(oldexe)

    for s in pe.sections:
        if s.Name == b".rsrc\0\0\0":
            rsrc_section = s
            break
    else:
        raise Exception("Couldn't find resource section.")

    base = rsrc_section.PointerToRawData
    resource_virtual = rsrc_section.VirtualAddress

    physize = rsrc_section.SizeOfRawData
    virsize = rsrc_section.Misc_VirtualSize

    f = open(oldexe, "rb")
    basedata = f.read(base)
    data = f.read(physize)

    #Symbol table, I could not understand so well
    #  Some analysis and the little I could find on how the symbol tables were layed out show that the PE table is in two sections. One is
    # a list of NumberOfSymbols*(18 bytes) symbol entries, followed immediately by a string table who's length is specified in the 4 bytes following the symbol structure list
    #Again, the information *seems* to indicate that this will always follow the entire block of all sections (so be at the end of the file)


    ######
    # As a related note, apparently this flag has been deprecated, but appears to be set (I think ..) in the renpy.exe that is used to build from..
    # no idea of a consequence, just an observation
    # https://docs.microsoft.com/en-us/windows/win32/debug/pe-format#characteristics , see "IMAGE_FILE_LINE_NUMS_STRIPPED" :
    #	COFF line numbers have been removed. This flag is deprecated and should be zero.
    # In renpy.exe, this 2-byte  flag is @ 0x096 and 0x097  in little-endian
    ######

    f.close()

    bf = BinFile(data)

    resources = parse_directory(bf, 0)
    # show_resources(resources, "")
    resources.update(load_icon(icofn))
    # show_resources(resources, "")

    rsrc = Packer().pack(resources)

    newExactSize = len(rsrc)

    #From note about flags above regarding image flags. These two should be 0
    pe.FILE_HEADER.Characteristics = pe.FILE_HEADER.Characteristics & (~pefile.IMAGE_CHARACTERISTICS["IMAGE_FILE_LINE_NUMS_STRIPPED"])
    pe.FILE_HEADER.Characteristics = pe.FILE_HEADER.Characteristics & (~pefile.IMAGE_CHARACTERISTICS["IMAGE_FILE_LOCAL_SYMS_STRIPPED"])

    #Per docs, symbol table can just be removed
    #https://docs.microsoft.com/en-us/windows/win32/debug/pe-format#coff-file-header-object-and-image  see "PointerToSymbolTable"  and  "NumberOfSymbols "

    pe.FILE_HEADER.PointerToSymbolTable = 0
    pe.FILE_HEADER.NumberOfSymbols = 0

    #####
    # Alignment requirements as I understand them
    #
    # SectionAlignment (misnomer.. 'memory' alignment) = 4096
    # FileAlignment = 512
    #
    #  Section, SizeOfRawData = multiple of FileAlignment   (On Disk size)
    #  Scetion, VirtualSize   is  NOT aligned to any specific number, and rwdata may be > than this due to its alignment requirement  (In memory size)
    #
    #  the resource section should be padded out to meet the alignment requirement of  SizeOfRawData = multiple of FileAlignment  (So, in the file, since on disk)
    #
    #####
    #  RVA table virt address and size = that in the section header
    #
    #####
    # SizeOfImage  must be aligned to SectionAlignment  (memory alignment)
    #   There is NO padding needed, the image size number simply needs to be aligned to >= the

    memoryalignment = pe.OPTIONAL_HEADER.SectionAlignment
    filealignment = pe.OPTIONAL_HEADER.FileAlignment

    rsrc_section.Misc_VirtualSize = newExactSize
    rsrc_section.Misc_PhysicalAddress = newExactSize
    rsrc_section.Misc = newExactSize

    #Size of resource section (and pointer, but we're not changing anything 'above' resource) is the same as virtual address in the section table
    # https://docs.microsoft.com/en-us/windows/win32/debug/pe-format#optional-header-data-directories-image-only
    pe.OPTIONAL_HEADER.DATA_DIRECTORY[2].Size = rsrc_section.Misc_VirtualSize


    #Alignment for raw data
    if len(rsrc) % filealignment:
        pad = filealignment - (len(rsrc) % filealignment)
        padding = b"\0" * (pad)
        padding = padding[:pad]
        rsrc += padding
    rsrc_section.SizeOfRawData = len(rsrc)


    #Image size is the memory size of all sections + all headers, padded to memory alignment.
    #There's alread a header size, and this isn't changing headers, so just add.
    #imageSize = pe.OPTIONAL_HEADER.ImageBase + pe.OPTIONAL_HEADER.SizeOfHeaders
    imageSize = pe.OPTIONAL_HEADER.SizeOfHeaders
    for s in pe.sections:
        sectionSize = s.Misc_VirtualSize
        #Align every section, so add padding to size
        pad = 0
        if sectionSize % memoryalignment:
            pad = memoryalignment - (sectionSize % memoryalignment)
        imageSize += (sectionSize + pad)
    pad = 0
    if imageSize % memoryalignment:
        pad = memoryalignment - (imageSize % memoryalignment)
    pe.OPTIONAL_HEADER.SizeOfImage = imageSize + pad


    #The symbol table is simply left off. Its size DOES factor into ImageSize, but we didn't calculate it above, so fine
    #Correctly checksum the file. The entire file is involved in the calculation, so a new PE object must be generated for the calculation to work against
    newFile = pe.write()[:base] + rsrc
    newpe = pefile.PE(data=bytes(newFile))
    newpe.OPTIONAL_HEADER.CheckSum = newpe.generate_checksum()

    return bytes(newpe.write())


if __name__ == "__main__":

    f = open(sys.argv[3], "wb")
    f.write(change_icons(sys.argv[1], sys.argv[2]))
    f.close()



# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function
from builtins import chr

include "linebreak.pxi"

cdef class Glyph:

    def __cinit__(self):
        self.variation = 0
        self.delta_x_offset = 0

    def __repr__(self):
        if self.variation == 0:
            return "<Glyph {0!r} time={1}>".format(self.character, self.time)
        else:
            return "<Glyph {0!r} vs={1} time={2}>".format(self.character, self.variation, self.time)

    _types = """
        x: int
        y: int
        delta_x_offset : int
        character : int
        variation : int
        split : int
        ruby : int
        ascent : int
        line_spacing : int
        width : float
        advance : float
        time : float
        hyperlink : int
        draw : bool
        """


cdef class Line:

    def __init__(self, int y, int height, list glyphs):
        self.y = y
        self.height = height
        self.glyphs = glyphs
        self.eop = False

    def __repr__(self):
        return "<Line y={0}, height={1}>".format(self.y, self.height)

    _types = """
        y : int
        height : int
        glyphs : list[Glyph]
        max_time : float
        eop : bool
        """


# The maximum width of text we lay out. This should be quite a bit smaller
# than the maximum SDL surface width. (16384)
cdef int MAX_WIDTH
MAX_WIDTH = 8192

TEXT=1
TAG=2
PARAGRAPH=3
DISPLAYABLE=4

def tokenize(unicode s):
    """
    This tokenizes a unicode string into text tags and tokens. It returns a list
    of pairs, where each pair begins with TEXT, TAG or PARAGRAPH, and then has
    the contents of the text run or tag.
    """

    cdef int TEXT_STATE = 1
    cdef int LEFT_BRACE_STATE = 2
    cdef int TAG_STATE = 3
    cdef int state = TEXT_STATE

    cdef Py_UCS4 c
    cdef unicode buf = u''

    cdef list rv = [ ]

    for c in s:

        if state == TEXT_STATE:

            if c == u'\n':
                if buf:
                    rv.append((TEXT, buf))

                rv.append((PARAGRAPH, u''))
                buf = u''

                continue

            elif c == u'{':
                state = LEFT_BRACE_STATE
                continue

            else:
                buf += c
                continue

        elif state == LEFT_BRACE_STATE:
            if c == u'{':
                buf += c
                state = TEXT_STATE
                continue

            elif c == u'}':
                raise Exception("Empty text tag in {0!r}.".format(s))

            else:
                if buf:
                    rv.append((TEXT, buf))

                buf = c
                state = TAG_STATE
                continue

        elif state == TAG_STATE:

            if c == u'}':
                rv.append((TAG, buf))
                buf = u''
                state = TEXT_STATE
                continue

            else:
                buf += c

    if state != TEXT_STATE:
        raise Exception("Open text tag at end of string {0!r}.".format(s))

    if buf:
        rv.append((TEXT, buf))

    return rv


def annotate_western(list glyphs):
    """
    Annotate the characters with line splitting information.
    """

    cdef Glyph g

    for g in glyphs:

        # Don't split ruby.
        if g.ruby != RUBY_NONE:
            continue

        if g.character == 0:
            g.split = SPLIT_BEFORE
        elif g.character == 0x20 or g.character == 0x200b:
            g.split = SPLIT_INSTEAD
        else:
            g.split = SPLIT_NONE


def annotate_anywhere(list glyphs):
    """
    allow all characters without ruby to be used for linebreaking.
    """

    cdef Glyph g

    for g in glyphs:

        # Don't split ruby.
        if g.ruby != RUBY_NONE:
            continue

        if g.character == 0x20 or g.character == 0x200b:
            g.split = SPLIT_INSTEAD
        else:
            g.split = SPLIT_BEFORE

# This is used to tailor the unicode break algorithm. If a character in this
# array is mapped to not
cdef char break_tailor[65536]

for i in range(0, 65536):
    break_tailor[i] = BC_XX

def language_tailor(chars, cls):
    """
    :doc: other
    :args: (chars, cls)

    This can be used to override the line breaking class of a character. For
    example, the linebreaking class of a character can be set to ID to
    treat it as an ideograph, which allows breaks before and after that
    character.

    `chars`
        A string containing each of the characters to tailor.

    `cls`
        A string giving a character class. This should be one of the classes defined in Table
        1 of `UAX #14: Unicode Line Breaking Algorithm <http://www.unicode.org/reports/tr14/tr14-30.html>`_.
    """

    ncls = CLASSES.get(cls, BC_XX)

    for c in chars:
        break_tailor[ord(c)] = ncls

# cjk
#
# 0 = western
# 1 = loose
# 2 = normal
# 3 = strict
def annotate_unicode(list glyphs, bint no_ideographs, int cjk):
    """
    Annotate unicode characters with information as to if they can be used
    for linebreaking.
    """

    cdef char old_type, new_type, tailor_type
    cdef int space_pos, pos
    cdef int len_glyphs = len(glyphs)
    cdef int c
    cdef char bc
    cdef Glyph g, g1, old_g
    cdef char *break_classes

    old_type = BC_WJ
    pos = 1
    space_pos = 0

    if not glyphs:
        return

    if cjk == 0:
        break_classes = break_western
    elif cjk == 1:
        break_classes = break_cjk_loose
    elif cjk == 2:
        break_classes = break_cjk_normal
    elif cjk == 3:
        break_classes = break_cjk_strict
    else:
        break_classes = break_western

    for pos from 1 <= pos < len_glyphs:

        g = glyphs[pos]
        c = g.character

        if 0x20000 <= c <= 0x2ffff: # Supplemental Ideographic Plane
            new_type = BC_ID
            tailor_type = BC_XX
        elif c > 65535: # Other non-basic planes.
            new_type = BC_AL
            tailor_type = BC_XX
        else: # Basic plane - use lookup table.
            new_type = break_classes[c]
            tailor_type = break_tailor[c]

        # If given no-ideographs, then turn ideographs and hangul syllables
        # into alphabetic characters.
        if no_ideographs and (
            new_type == BC_H2 or
            new_type == BC_H3 or
            new_type == BC_ID or
            new_type == BC_JL or
            new_type == BC_JV or
            new_type == BC_JT):

            new_type = BC_AL

        # Normalize the class by turning various groups into AL.
        if (new_type >= BC_PITCH and new_type != BC_SP and new_type != BC_CB):
            new_type = BC_AL

        if tailor_type != BC_XX:
            new_type = tailor_type

        # If we have a space, record it and continue.
        if new_type == BC_SP:
            g.split = SPLIT_NONE
            space_pos = pos
            continue

        # If we have a combining mark, continue.
        if new_type == BC_CM:
            g.split = SPLIT_NONE
            continue

        if new_type == BC_CB:
            if old_type == BC_WJ or old_type == BC_GL:
                g.split = SPLIT_NONE
            else:
                g.split = SPLIT_BEFORE

            continue

        if old_type == BC_CB:
            if new_type == BC_WJ or new_type == BC_GL:
                g.split = SPLIT_NONE
            else:
                g.split = SPLIT_BEFORE

            continue

        # Figure out the type of break opportunity we have here.
        # ^ Prohibited break.
        # % Indirect break.
        # @ Prohibited break (combining mark)
        # # Indirect break (combining mark)
        # _ Direct break.

        bc = break_rules[ old_type * BC_PITCH + new_type]

        if bc == "%": # Indirect break.
            if space_pos:
                g1 = glyphs[space_pos]
                g1.split = SPLIT_INSTEAD

            g.split = SPLIT_NONE

        elif bc == "_": # Direct break.
            if space_pos:
                g1 = glyphs[space_pos]
                g1.split = SPLIT_INSTEAD
                g.split = SPLIT_NONE
            else:
                g.split = SPLIT_BEFORE

        else:
            g.split = SPLIT_NONE

        old_type = new_type
        space_pos = 0

    # Deal with ruby, by marking it as non-spacing.
    old_g = glyphs[0]
    old_g.split = SPLIT_NONE

    for g in glyphs:

        if g.character == 0:
            g.split = SPLIT_BEFORE

        if g.ruby == RUBY_TOP or g.ruby == RUBY_ALT:
            g.split = SPLIT_NONE

        elif g.ruby == RUBY_BOTTOM and old_g.ruby == RUBY_BOTTOM:
            g.split = SPLIT_NONE

        old_g = g



def linebreak_greedy(list glyphs, int first_width, int rest_width):
    """
    Starting with a list of glyphs, decides where to split it. The result of
    this is the same list of glyphs, but with some of the .split fields set
    back to SPLIT_NONE, where we decided not to go ahead with the split after
    all.

    This is the greedy algorithm, which splits when the line is longer
    than a specified width.
    """

    cdef Glyph g, split_g
    cdef float width, x, splitx, gwidth

    width = first_width
    split_g = None

    # The x position of the current character. Invariant: x can never be more
    # that one character-width greater than width.
    x = 0

    # The x position after splitting the line.
    splitx = 0

    for g in glyphs:

        if g.ruby == RUBY_TOP:
            continue

        if g.ruby == RUBY_ALT:
            continue

        # If the x coordinate is greater than the width of the screen,
        # split at the last split point, if any.
        if x > width and split_g is not None:
            x = splitx
            split_g = None
            width = rest_width

        x += g.advance
        splitx += g.advance

        if g.split == SPLIT_INSTEAD:
            if split_g is not None:
                split_g.split = SPLIT_NONE

            split_g = g
            splitx = 0

        elif g.split == SPLIT_BEFORE:

            if split_g is not None:
                split_g.split = SPLIT_NONE

            split_g = g
            splitx = g.advance

    # Split at the last character, if necessary.
    if x > width:
        split_g = None

    if split_g is not None:
        split_g.split = SPLIT_NONE

def linebreak_nobreak(list glyphs):
    """
    Linebreak without linebreaking.
    """

    cdef Glyph g

    for g in glyphs:
        g.split = SPLIT_NONE


def linebreak_debug(list glyphs):
    """
    Return a string giving the results of linebreaking a list of glyphs.
    """

    cdef Glyph g

    rv = ""

    for g in glyphs:

        if g.split == SPLIT_INSTEAD:
            rv += "|"
        elif g.split == SPLIT_BEFORE:
            rv += "[" + chr(g.character)
        else:
            rv += chr(g.character)

    return rv


def linebreak_list(list glyphs):
    """
    Returns a list of unicode strings, one per broken line.
    """

    cdef Glyph g

    rv = [ ]
    line = u""

    for g in glyphs:

        if g.split == SPLIT_INSTEAD:
            rv.append(line)
            line = u""
        elif g.split == SPLIT_BEFORE:
            rv.append(line)
            line = chr(g.character)
        else:
            line += chr(g.character)

    if line:
        rv.append(line)

    return rv


def place_horizontal(list glyphs, float start_x, float first_indent, float rest_indent):
    """
    Place the glyphs horizontally, without taking into account the indentation
    at the start of the line. Returns the width of the laid-out line.
    """

    if not glyphs:
        return 0

    cdef Glyph g, old_g
    cdef float x, maxx

    x = start_x + first_indent
    maxx = 0
    old_g = None

    for g in glyphs:

        if g.ruby == RUBY_TOP:
            continue

        if g.ruby == RUBY_ALT:
            continue

        if g.split != SPLIT_NONE and old_g:
            # When a glyph is at the end of the line, set its advance to
            # be its width. (This makes things like strikeout and underline
            # easier, since we only need consider advance.)
            old_g.advance = old_g.width

        if g.split == SPLIT_INSTEAD:
            x = start_x + rest_indent
            continue

        elif g.split == SPLIT_BEFORE:
            x = start_x + rest_indent

        g.x = <short> (x + .5)

        if maxx < x + g.width:
            maxx = x + g.width
        if maxx < x + g.advance:
            maxx = x + g.advance

        x += g.advance

        # Limit us to some width.
        if x > MAX_WIDTH:
            x = MAX_WIDTH

        old_g = g

    return maxx

def place_vertical(list glyphs, int y, int spacing, int leading):
    """
    Vertically places the non-ruby glyphs. Returns a list of line end heights,
    and the y-value for the top of the next line.
    """

    cdef Glyph g, gg

    cdef int pos, sol, len_glyphs, i
    cdef int ascent, line_spacing
    cdef bint end_line

    if not glyphs:
        return [ ], y

    len_glyphs = len(glyphs)

    pos = 0
    sol = 0

    ascent = 0
    line_spacing = 0

    rv = [ ]

    y += leading

    while True:

        if pos >= len_glyphs:
            end_line = True
        else:
            g = glyphs[pos]
            end_line = (g.split != SPLIT_NONE)

        if end_line:

            for i from sol <= i < pos:
                gg = glyphs[i]

                if gg.ruby == RUBY_TOP:
                    continue

                if gg.ruby == RUBY_ALT:
                    continue

                if gg.ascent:
                    gg.y = y + ascent

                else:
                    # Glyphs without ascents are displayables, which get
                    # aligned to the top of the line.
                    gg.y = y
                    gg.ascent = ascent

            l = Line(y - leading, leading + line_spacing + spacing, glyphs[sol:pos])
            rv.append(l)

            y += line_spacing
            y += spacing
            y += leading

            sol = pos

            ascent = 0
            line_spacing = 0

            if g.split == SPLIT_INSTEAD:
                sol += 1
                pos += 1

                if pos >= len_glyphs:
                    break
                else:
                    continue

        if pos >= len_glyphs:
            break

        if g.ascent > ascent:
            ascent = g.ascent

        if g.line_spacing > line_spacing:
            line_spacing = g.line_spacing

        pos += 1


    rv[-1].eop = True
    return rv, y - leading

def kerning(list glyphs, float amount):
    cdef Glyph g

    for g in glyphs:
        g.advance += amount


def assign_times(float t, float gps, list glyphs):
    """
    Assign a display time to each glyph.

    `t`
        The start time of the first glyph.

    `gps`
        The number of glyphs per second to show.

    `glyphs`
        A list of glyphs to apply this to.

    Returns the time of the first glyph in the next block.
    """

    cdef float tpg # time per glyph
    cdef Glyph g

    if gps == 0:
        tpg = 0.0
    else:
        tpg = 1.0 / gps

    for g in glyphs:

        if (g.ruby == RUBY_TOP) or (g.ruby == RUBY_ALT):
            g.time = -1
            continue

        t += tpg
        g.time = t


    return t


def max_times(list l):
    """
    Set the max_time filed on each line.
    """

    cdef Line line
    cdef Glyph g
    cdef float max_time

    max_time = 0

    for line in l:
        for g in line.glyphs:
            if g.time > max_time:
                max_time = g.time

        line.max_time = max_time

    return max_time


def hyperlink_areas(list l):
    """
    Returns a list of (hyperlink, x, y, w, h) tuples, where each entry in
    the rectangle represents a contiguous portion of a hyperlink on the
    given line.
    """

    cdef Line line
    cdef Glyph g
    cdef list gl
    cdef int len_gl

    cdef int pos

    cdef int max_x
    cdef int min_x
    cdef int hyperlink

    rv = [ ]

    for line in l:
        gl = line.glyphs
        len_gl = len(gl)

        hyperlink = 0
        max_x = 0
        min_x = 1000000
        pos = 0

        while pos < len_gl:

            g = gl[pos]

            if (hyperlink and g.hyperlink != hyperlink):
                rv.append((hyperlink, min_x, line.y, max_x - min_x, line.height))
                hyperlink = 0
                max_x = 0
                min_x = 1000000

            hyperlink = g.hyperlink

            if hyperlink:
                if g.x < min_x:
                    min_x = g.x

                if g.x + g.width > max_x:
                    max_x = g.x + <int> g.width

            pos += 1

        if hyperlink:
            rv.append((hyperlink, min_x, line.y, max_x - min_x, line.height))

    return rv


def mark_ruby_top(list l):

    cdef Glyph g

    for g in l:
        g.ruby = RUBY_TOP

def mark_altruby_top(list l):

    cdef Glyph g

    for g in l:
        g.ruby = RUBY_ALT


def mark_ruby_bottom(list l):

    cdef Glyph g

    for g in l:
        g.ruby = RUBY_BOTTOM


def place_ruby(list glyphs, int ruby_offset, int altruby_offset, int surf_width, int surf_height):

    cdef Glyph g
    cdef ruby_t last_ruby = RUBY_NONE
    cdef int len_glyphs = len(glyphs)
    cdef float x, width, min_x = 0, max_x = 0
    cdef int y = 0
    cdef int start_top
    cdef int pos = 0
    cdef int i

    while pos < len_glyphs:

        g = glyphs[pos]

        if g.ruby == RUBY_NONE:

            min_x = g.x
            max_x = g.x + g.width
            y = g.y

            last_ruby = RUBY_NONE

            pos += 1
            continue

        elif g.ruby == RUBY_BOTTOM:

            if last_ruby != RUBY_BOTTOM:
                min_x = g.x

            max_x = g.x + g.width
            y = g.y

            last_ruby = RUBY_BOTTOM

            pos += 1
            continue

        # Otherwise, we have RUBY_TOP or RUBY_ALT

        # Find the run of RUBY_TOP. When this is done, the run will be in
        # glyphs[start_top:pos].
        start_top = pos

        last_ruby = g.ruby

        while pos < len_glyphs:

            g = glyphs[pos]
            if g.ruby != last_ruby:
                break

            pos += 1

        # Compute the width of the run.

        width = 0
        for i from start_top <= i < pos:
            g = glyphs[i]
            width += g.advance

        width -= glyphs[pos - 1].advance
        width += glyphs[pos - 1].width

        # Place the glyphs.
        x = (max_x + min_x) / 2 - width / 2

        for i from start_top <= i < pos:
            g = glyphs[i]
            g.x = <int> (x + .5)

            if g.ruby == RUBY_TOP:
                g.y = y + ruby_offset
            else:
                g.y = y + altruby_offset

            x += g.advance


def align_and_justify(list lines, int width, float text_align, bint justify):
    """
    Handle text alignment and justification.
    """

    cdef Line l
    cdef Glyph g

    cdef int max_x
    cdef int spaces

    cdef float justify_offset
    cdef float justify_per_space

    cdef int offset

    # See if we have to do anything at all.
    if not justify and text_align == 0.0:
        return

    for l in lines:

        spaces = 0
        max_x = 0

        for g in l.glyphs:

            if g.ruby == RUBY_TOP:
                continue

            if g.ruby == RUBY_ALT:
                continue

            if g.character == 0x20:
                spaces += 1

            max_x = <int> (g.x + g.width)

        # If we're too big, give up.
        if max_x >= MAX_WIDTH:
            continue

        if justify and spaces and not l.eop:

            justify_per_space = 1.0 * (width - max_x) / spaces
            justify_offset = 0.5 # Makes numbers round better.

            for g in l.glyphs:

                if g.ruby == RUBY_TOP:
                    continue

                if g.ruby == RUBY_ALT:
                    continue

                if g.character == 0x20:
                    justify_offset += justify_per_space

                g.x += <int> justify_offset

        else:
            offset = <int> ((width - max_x) * text_align)

            for g in l.glyphs:

                if g.ruby == RUBY_TOP:
                    continue

                if g.ruby == RUBY_ALT:
                    continue

                g.x += offset

def reverse_lines(list glyphs):
    """
    Reverses each line in glyphs, while keeping the lines themselves in
    the original order.
    """

    cdef list rv
    cdef list block
    cdef Glyph g

    rv = [ ]
    block = [ ]

    for g in glyphs:

        if g.split == SPLIT_INSTEAD:
            block.reverse()
            rv.extend(block)
            rv.append(g)
            block = [ ]

            continue

        block.append(g)

    block.reverse()
    rv.extend(block)

    return rv

def copy_splits(list source, list dest):
    """
    Copies break and timing information from one list of glyphs
    to another.
    """

    cdef Glyph s
    cdef Glyph d
    cdef int i

    for 0 <= i < len(dest):
        s = source[i]
        d = dest[i]

        d.split = s.split


def tweak_glyph_spacing(list glyphs, list lines, double dx, double dy, double w, double h):
    cdef Glyph g

    if w <= 0 or h <= 0:
        return

    cdef short old_x_offset = 0
    cdef short x_offset

    for g in glyphs:

        x_offset = <short> (dx * g.x / w)

        g.x += x_offset
        g.y += <short> (dy * g.y / h)

        if x_offset > old_x_offset:
            g.delta_x_offset = x_offset - old_x_offset

        old_x_offset = x_offset

    for l in lines:
        end = l.y + l.height

        l.y += int(dy * l.y / h)
        end += int(dy * end / h)

        l.height = end - l.y

def offset_glyphs(list glyphs, short x, short y):
    cdef Glyph g

    if x == 0 and y == 0:
        return

    for g in glyphs:
        g.x += x
        g.y += y

"This exists to force a recompile for Ren'Py 8.0.2 and 7.5.2."


# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# This file contains code to emulate various other devices on the PC.

from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *


import os

import pygame_sdl2 as pygame
import renpy


# The function that's called to perform the emulation. This function has
# the signature of null_emulator.
emulator = None

# An overlay that is placed over the screen to support the emulator.
overlay = [ ]

# True if we're in ios mode, where we don't allow keys.
ios = False


def null_emulator(ev, x, y):
    """
    This is used when emulation is not desired.
    """
    return ev, x, y


TOUCH_KEYS = [ pygame.K_ESCAPE, pygame.K_PAGEUP ]


def touch_emulator(ev, x, y):
    """
    This emulates an android touch-screen device, like a tablet or smartphone.
    """

    if renpy.display.interface.text_rect:
        return ev, x, y

    if ev.type == pygame.MOUSEBUTTONDOWN:
        if ev.button != 1:
            return None, x, y

    elif ev.type == pygame.MOUSEBUTTONUP:
        if ev.button != 1:
            return None, x, y

        move = pygame.event.Event(pygame.MOUSEMOTION, { "pos" : (0, 0), "rel" : (0, 0), "buttons" : (0, 0, 0) })
        renpy.display.interface.pushed_event = move

    elif ev.type == pygame.MOUSEMOTION:
        if not ev.buttons[0]:
            x = 0
            y = 0

    elif ev.type == pygame.KEYDOWN and not ios:
        if ev.mod & pygame.KMOD_SHIFT:
            pass
        elif not ev.key in TOUCH_KEYS:
            return None, x, y

    elif ev.type == pygame.KEYUP and not ios:
        if ev.mod & pygame.KMOD_SHIFT:
            pass
        if not ev.key in TOUCH_KEYS:
            return None, x, y

    return ev, x, y


TV_KEYS = [ pygame.K_LEFT, pygame.K_RIGHT, pygame.K_UP, pygame.K_DOWN, pygame.K_RETURN, pygame.K_ESCAPE, pygame.K_PAGEUP ]


def tv_emulator(ev, x, y):
    """
    This emulates a tv-based device, like the OUYA.
    """

    if ev.type == pygame.MOUSEBUTTONDOWN:
        return None, x, y
    elif ev.type == pygame.MOUSEBUTTONUP:
        return None, x, y
    elif ev.type == pygame.MOUSEMOTION:
        return None, x, y
    elif ev.type == pygame.KEYDOWN:
        if not ev.key in TV_KEYS:
            return None, x, y
    elif ev.type == pygame.KEYDOWN:
        if not ev.key in TV_KEYS:
            return None, x, y

    return ev, x, y


keyboard = None
null = None


def dynamic_keyboard(st, at):
    global keyboard
    global null

    if keyboard is None:
        keyboard = renpy.store.Fixed(
            renpy.store.Solid("#0008", yalign=1.0, ymaximum=.625),
            renpy.store.Text("On-Screen Keyboard", xalign=.5, yalign=.75),
            )
        null = renpy.store.Null()

    if renpy.display.interface.old_text_rect:
        rv = keyboard
    else:
        rv = null

    return rv, .33


def init_emulator():
    """
    Sets up the emulator.
    """

    global emulator
    global overlay
    global ios

    name = os.environ.get("RENPY_EMULATOR", "")

    if name == "touch":
        emulator = touch_emulator
        overlay = [ renpy.store.DynamicDisplayable(dynamic_keyboard) ]

    elif name == "ios-touch":
        emulator = touch_emulator
        overlay = [ renpy.store.DynamicDisplayable(dynamic_keyboard) ]
        ios = True

    elif name == "tv":
        emulator = tv_emulator
        overlay = [ renpy.display.motion.Transform(
            "_tv_unsafe.png",
            xalign=0.5,
            yalign=0.5,
            size=(int(renpy.config.screen_height * 16.0 / 9.0), renpy.config.screen_height),
            ) ]

    else:
        emulator = null_emulator
        overlay = [ ]


def early_init_emulator():

    name = os.environ.get("RENPY_EMULATOR", "")

    if name:

        renpy.exports.windows = False # type: ignore
        renpy.exports.linux = False # type: ignore
        renpy.exports.macintosh = False # type: ignore
        renpy.exports.web = False # type: ignore
        renpy.exports.android = renpy.exports.variant("android") # type: ignore
        renpy.exports.ios = renpy.exports.variant("ios") # type: ignore
        renpy.exports.mobile = renpy.exports.android or renpy.exports.ios  # type: ignore
